import re
import time
import logging
import subprocess
import ConfigParser
import numpy as np
import socket
from ... import config as kcg_config

log = logging.getLogger(__name__)

HEADER_SIZE_BYTES = 32

class BoardError(Exception):
    pass

class InterfaceNotFoundError(Exception):
    pass

class ObserverError(Exception):
    pass

class BoardConfiguration():

    def __init__(self, config_file=None, logger=None):
        self._config = {}
        self._observers = {}
        self.logger = logger
        self._set_defaults()
        self.set_default_observers()


        if config_file:
            self.load_config(config_file)

    def _set_defaults(self):
        c = self._config
        c['fpga_delay_max']       = 15
        c['fpga_delay']           = 0
        c['fpga_delay_factor']    = 150
        
        c['chip_delay_max']       = 31
        c['chip_1_delay']         = 4
        c['chip_2_delay']         = 4
        c['chip_3_delay']         = 4
        c['chip_4_delay']         = 4
        c['chip_delay_factor']    = 3

        c['th_delay_max']         = 15
        c['th_delay']             = 3
        c['th_delay_factor']      = 150
        
        c['adc_delay_max']        = 15
        c['adc_1_delay']          = 4
        c['adc_2_delay']          = 4
        c['adc_3_delay']          = 4
        c['adc_4_delay']          = 4
        c['adc_delay_factor']     = 150
        
        c['th_to_adc_cycles']     = 7
        c['adc_1_delay_individual'] = -1  #      -1 = 'ADC 1 delay is the same as th_to_adc_cycles'
                                          # 0 .. 16 = 'Use this offset instead'
        
        c['orbits_observe']       = 100
        c['orbits_skip']          = 2
        c['acquisition_count']    = 10
        c['orbits_wait_time']     = 15
        c['trigger_skip']         = 0
        c['trigger_timeout']      = 12
        c['trigger_method']       = 1
        c['use_trigger']          = False
        c['build_spectrograms']   = False
        c['pilot_bunch']          = False
        if kcg_config.save_header == True:  # Do not write directly in case of damaged config file
            c['header']           = True
        else:
            c['header']           = False

    def set_default_observers(self):
        self.observe(None, self.update_header, 'header')
        self.observe(None, lambda x: write_pci(hex(x), '0x9020'), 'orbits_observe')
        self.observe(None, lambda x: write_pci(hex(x), '0x9028'), 'orbits_skip')


    def notify_all_observers(self):
        for key in self._config.keys():
            observers = self._observers.get(key, None)
            if observers:
                for (who, callback) in observers:
                    callback(self.get(key))



    def register_logger(self, logger):
        self.logger = logger

    def load_config(self, filename):
        if filename:
            config = ConfigParser.RawConfigParser()
            if not config.read(str(filename)):
                return False

            for key in self._config.keys():
                try:
                    if type(self._config[key]) == int:
                        self._config[key] = config.getint('Config', key)
                    elif type(self._config[key]) == bool:
                        self._config[key] = config.getboolean('Config', key)
                    if self.logger:
                        self.logger.info("Read '%s' for '%s' from '%s'"%(str(self._config[key]), key, str(filename)))
                except ConfigParser.NoOptionError as e:
                    pass
                except ConfigParser.NoSectionError as e:
                    pass
            return True
        else:
            return False

    def save_config(self, filename):
        if filename:
            # with open(filename, 'w') as f:
            try:
                f = open(filename, 'w')
                cp = ConfigParser.RawConfigParser()
                cp.add_section('Config')
                for key in self._config.keys():
                    cp.set('Config', key, self._config[key])
                f.write('#\n'
                        '#  KCG   (KAPTURE Control Gui) Configuration file\n'
                        '#\n'
                        '#  (c) Karlsruhe Institute of Technology, 2015\n'
                        '#  All rights reserved.\n'
                        '#\n'
                        '#  Applicable Gui Version(s): 1.0 - 1.0.2\n'
                        '#\n'
                        '#  Saved at: ' + time.asctime() + '\n'
                        '#\n\n')
                cp.write(f)
            except (IOError, TypeError):
                return False
            return True
        else:
            return False

    def get(self, key):
        return self._config.get(key, None)

    def dump(self):
        s = ""
        for key in self._config.keys():
            s += key + ": " + str(self.get(key)) + ", "
        return s[:-1]

    def update(self, key, value):
        self._config[key] = value
        self._notify_observers(key, value)
    def updateSilent(self, key, value):
        self._config[key] = value

    def observe(self, who, callback, key):
        if key not in self._config.keys():
            raise ObserverError(str("Key '%s' is unknown."%key))
        
        if self._observers.get(key, None) is None:
            self._observers[key] = []
        
        self._observers[key].append([who, callback])
        
    def unobserve(self, who, key=None):
        if key is not None:
            observers = np.array(self._observers.get(key, None))
            if observers is None:
                return
            if who not in observers[:, 0]:
                return               
            for i, _obs in enumerate(self._observers[key]):
                if _obs[0] is who:
                    del self._observers[key][i]
                    if not self._observers[key]:
                        del self._observers[key]
            return
        
        for _key in self._observers.keys():
            for i, _obs in enumerate(self._observers[_key]):
                if _obs[0] is who:
                    del self._observers[_key][i]
                    if not self._observers[_key]:
                        del self._observers[_key]

    def _notify_observers(self, key, value):
        observers = self._observers.get(key, None)
        if observers is None:
            return
        for (who, callback) in observers:
            callback(value)
        
    def make_uint(self, value, maximum, name=None):
        if value is None:
            raise ValueError(str("%s Value is invalid (None)" % name))
        
        val = None
        try:
            val = int(value)
        except ValueError:
            raise ValueError(str("%s Value is not a valid number" % name))

        if maximum is not None:    
            if val > maximum:
                raise ValueError(str("%s Value is too large (>%i)" % (name, maximum)))
                
        if val < 0:
            raise ValueError(str("%s Values below 0 are not allowed" % name))
        
        return val

    def set_fpga_delay(self, value):
        time_factor = self.make_uint(value, self.get('fpga_delay_max'), 'FPGA_Delay')
        reg_value = "0x000501" + '{0:01x}'.format(time_factor) + "0"
        write_pci(reg_value, '0x9060')
        if self.logger:
            self.logger.info("Set FPGA clock delay %i * %i --> %i picoseconds" % (time_factor, self.get('fpga_delay_factor'), time_factor*self.get('fpga_delay_factor')))
        self.update('fpga_delay', value)

    def set_chip_delay(self, adcs, values):
        if not adcs or not values:
            if self.logger:
                self.logger.info("Nothing to do for chip delay.")
            return

        _adcs = []
        for adc in adcs:
            _adcs.append(self.make_uint(adc, 3, 'ADC_'))
            
        _values = []
        for value in values:
            _values.append(self.make_uint(value, self.get('chip_delay_max'), 'ADC_Value'))

        a_v = zip(_adcs, _values) 
            
        factors = [None, None, None, None]
        for (adc, value) in a_v:
            factors[adc] = value
            
        reg_value = ''
        mask = ''
        # Chip Delays are stored as 'ADC_4 ADC_3 ADC_2 ADC_1' in the register.
        # Therefore, we need to traverse the factors array in reverse order
        for value in reversed(factors):
            if value is not None:
                reg_value += '{0:02x}'.format(value)
                mask += 'ff'
            else:
                reg_value += '00'
                mask += '00'
               
        write_pci(reg_value, '0x9080', hex_mask=mask)
        if self.logger:
            s = "Setting ADC Delays:"
            for (adc, value) in a_v:
                s += ' ADC_%i Fine Delay: %i,' % (adc, value)
            s = s[:-1]  # cut away the last dangling ','
            self.logger.info(s)
        for (adc, value) in a_v:
            s = 'chip_%i_delay'%(adc+1)
            self.update(s, value)

    def set_th_delay(self, value):
        time_factor = self.make_uint(value, self.get('th_delay_max'), 'TH_Delay')
        reg_value = "0x000501" + '{0:01x}'.format(time_factor) + "3"
        write_pci(reg_value, '0x9060')
        if self.logger:
            self.logger.info("Set T/H Signal delay %i * %i --> %i picoseconds" % (time_factor, self.get('th_delay_factor'), time_factor*self.get('th_delay_factor')))
        self.update('th_delay', value)

    def set_adc_delay(self, adc, value):
        if adc is None or value is None:
            if self.logger:
                self.logger.info("Nothing to do for ADC delay.")
            return
        _adc = self.make_uint(adc, 3, 'ADC Number')    
        _val = self.make_uint(value, self.get('adc_delay_max'), 'ADC Delay')
        reg_value = "0x000501" + '{0:01x}'.format(_val) + "%i" % (_adc+4)
        write_pci(reg_value, '0x9060')
        if self.logger:
            s = "Setting ADC_%i delay %i * %i --> %i picoseconds" % ((_adc+1), _val, self.get('adc_delay_factor'), _val*self.get('adc_delay_factor'))
            self.logger.info(s)
        adc_s = 'adc_%i_delay'%(_adc+1)
        self.update(adc_s, _val)

    def set_delay(self, n, ignore_seperate_delay=False):
        def write_delay(value, channel):
            cmd = '00501' + '%01x' % value + str(channel)
            write_pci(cmd, reg='0x9060')
            time.sleep(0.005)

        self.logger.info("Setting T/H Delay: " + str(n))
        write_delay(n, 3)
        self.update('th_delay', n)

        delay = n + self.get('th_to_adc_cycles')

        if delay > self.get('adc_delay_max'):
            delay -= self.get('adc_delay_max') + 1

        write_delay(delay, 5)
        self.update('adc_2_delay', delay)
        write_delay(delay, 6)
        self.update('adc_3_delay', delay)
        write_delay(delay, 7)
        self.update('adc_4_delay', delay)

        #ADC 1 might have an individual delay
        if self.get('adc_1_delay_individual') > 0:
            try:
                delay = n + self.make_uint(self.get('adc_1_delay_individual'), 16, 'ADC 1 individual delay')
                if self.logger:
                    self.logger.info("Setting ADC1 individual delay to " + str(delay))
            except ValueError:
                if self.logger:
                    self.logger.info(r"'adc_1_delay_individual' not set or inactive. Using default.")

            if delay > self.get('adc_delay_max'):
                delay -= self.get('adc_delay_max') + 1
        write_delay(delay, 4)
        self.update('adc_1_delay', delay)

    def update_header(self, state):
        """
        Set the flag to write Header to files when acquiring.
        :param state: True to enabling header and False to disable
        :return: -
        """
        try:
            control = read_pci(1, '0x9040')[0]
            control_bits = '{0:032b}'.format(int(control, 16))
            if state:
                control_bits = control_bits[:3] + '1' + control_bits[4:]
            else:
                control_bits = control_bits[:3] + '0' + control_bits[4:]
            dec_val_bits = int(control_bits, 2)
            write_pci(hex(dec_val_bits), '0x9040')
        except BoardError as e:
            reason = str(e) if str(e) != '' else "Unknown"
            self.logger.info("Error in Board Communication, was unable to write value to board "+reason)

    def read_from_board(self):
        settings = ['chip_1_delay','chip_2_delay','chip_3_delay','chip_4_delay']
        # --[ read fine/chip delays ]
        val = read_pci(reg='9080')[0]
        # --[ set chip_1_delay ]--
        self.update('chip_1_delay', int(val[6:8], 16))
        # --[ set chip_2_delay ]--
        self.update('chip_2_delay', int(val[4:6], 16))
        # --[ set chip_3_delay ]--
        self.update('chip_3_delay', int(val[2:4], 16))
        # --[ set chip_4_delay ]--
        self.update('chip_4_delay', int(val[0:2], 16))

        # --[ read and set th delay ]--
        val = read_pci(reg='90a0')[0]
        self.update('th_delay', int(val, 16))

        # --[ check for seperate adc1 delay ]--
        val = read_pci(reg='9088')[0]
        if int(val, 16) != self.get('th_delay') + self.get('adc_1_delay_individual'):
            self.update('adc_1_delay_individual', int(val, 16)-self.get('th_delay'))
        else:
            self.update('adc_1_delay_individual', -1)

        # --[ read and set number of orbits to acquire ]--
        val = read_pci(reg='9020')[0]
        self.update('orbits_observe', int(val, 16))

        # --[ read and set number of orbits to skip ]--
        val = read_pci(reg='9028')[0]
        self.update('orbits_skip', int(val, 16))

        # --[ read and update header ]--
        control = read_pci(1, '0x9040')[0]
        control_bits = '{0:032b}'.format(int(control, 16))
        if control_bits[3] == '1':
            self.update('header', True)
        else:
            self.update('header', False)

    @property
    def contiuous_read(self):
        return self.continu

config = BoardConfiguration()

def safe_call(cmd):
    log.debug(cmd)
    try:
        # if '-r' in cmd:
        return subprocess.check_output(cmd)
        # else:
        #     subprocess.Popen(cmd, shell=False)
    except subprocess.CalledProcessError as e:
        raise BoardError(e.output)
    except OSError as e:
        if str(e) == "[Errno 2] No such file or directory":
            raise InterfaceNotFoundError("pci command was not found in your searchpath")
        else:
            raise BoardError('{}: {}'.format(' '.join(cmd), str(e)))

backup_safe_call = safe_call

host = None
port = None
def safe_remote_call(cmd):
    soc = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    soc.connect((host, port))
    soc.send(str(cmd))
    soc.settimeout(None)
    l = int(soc.recv(100))
    print l
    ret = ""
    for i in range(1, l/4096 + 1):
        ret += soc.recv(i*4096)
    ret += soc.recv(l%4096)
    soc.close()
    print ret
    return ret

def write_pci(value, reg='0x9040', opts=[], hex_mask='FFFFFFFF'):
    if hex_mask != 'FFFFFFFF':
        if len(hex_mask) > 8:
            hex_mask = hex_mask[-8:]
        prev_val = read_pci(1, reg)[0]
        prev_bits = '{0:032b}'.format(int(prev_val,16))
        mask_bits = '{0:032b}'.format(int(hex_mask,16))
        value_bits = '{0:032b}'.format(int(value,16))
        new_bits = list(prev_bits)
        
        for i, bit in enumerate(mask_bits):
            if bit == '1':
                new_bits[i] = value_bits[i]
                
        value = hex(int("".join(new_bits),2))

    cmd = ['pci', '-w', reg, value]
    cmd.extend(opts)
    safe_call(cmd)
    log.debug('Written %str to register %s'%(value, reg))
    

def read_pci(amount=1, reg='0x9000', opts=[], decimal=False):
    cmd = ['pci', '-r', reg, "-s%i"%amount]
    cmd.extend(opts)
    output = safe_call(cmd).split("\n")
    lines = []
    for line in output:
        if line and line != "\n":
            lines.append(line.split(":  ")[1])
    formated = []
    for line in lines:
        if line and line != "\n":
            formated += re.findall(r'[\da-fA-F]{8}', line) 
    if decimal:
        return [int(x,16) for x in formated]
    else:
        return formated

    
def get_dec_from_bits(bits, msb=-1, lsb=-1):
    rtrnValue = 0
    
    if (msb < 0 or lsb < 0):
        rtrnValue = int(bits)
    elif (msb < lsb) or (msb > 31) or (lsb > 31):
        log.info("Bit range for msb and lsb of get_dec_from_bits was wrong. Not truncating")
        rtrnValue = int(bits, 2)
    else:
        chunk = ('{0:032b}'.format(int(bits, 2)))[31-msb:32-lsb]
        rtrnValue = int(chunk, 2)
        
    return rtrnValue

def get_status():
    registers = read_pci(3, '0x9050', decimal=True)
    bits = []
    bits += ['{0:032b}'.format(registers[0])]
    bits += ['{0:032b}'.format(registers[1])]
    bits += ['{0:032b}'.format(registers[2])]

    status = {}

    s1 = get_dec_from_bits(bits[0], 2, 0)
    if s1 == 0:
        # Pipeline in reset mode
        # self.pipeline_led.set_tri()
        status['pipeline'] = 2
    elif s1 == 1:
        # Pipeline is idle
        # self.pipeline_led.set_on()
        status['pipeline'] = 3
    elif s1 == 6:
        # Pipeline in error state
        # self.pipeline_led.set_off()
        status['pipeline'] = 1
    else:
        # Should not happen!
        # self.pipeline_led.set_out()
        status['pipeline'] = 0

    s2 = get_dec_from_bits(bits[0], 29, 26)
    if s2 == 0:
        # Master Control in reset mode
        # self.master_control_led.set_tri()
        status['master_control'] = 2
    elif s2 == 1:
        # Master Control is idle
        # self.master_control_led.set_on()
        status['master_control'] = 3
    elif s2 == 8:
        # Master Control in error state
        # self.master_control_led.set_off()
        status['master_control'] = 1
    else:
        # Should not happen!
        # self.master_control_led.set_out()
        status['master_control'] = 0

    s3 = get_dec_from_bits(bits[2], 15, 12)
    if s3 == 15:
        # Data Check Idle
        # self.data_check_led.set_on()
        status['data_check'] = 3
    else:
        # Data Check Error
        # self.data_check_led.set_off()
        status['data_check'] = 1

    s4 = int(bits[0][7])
    if s4 == 0:
        # PLL_LD not active
        # self.pll_ld_led.set_tri()
        status['PLL_LD'] = 2
    elif s4 == 1:
        # PLL_LD is active
        # self.pll_ld_led.set_on()
        status['PLL_LD'] = 3
    else:
        status['PLL_LD'] = 0

    return status



def start_dma(dma='dma0r'):
    log.info('Start DMA')
    cmd = ['pci', '--start-dma', dma]
    safe_call(cmd)
    time.sleep(0.05)
    

def stop_dma(dma='dma0r'):
    log.info('Stop DMA')
    cmd = ['pci', '--stop-dma', dma]
    safe_call(cmd)
    time.sleep(0.05)


def data_reset():
    log.info('Data reset')
    write_pci('000003f5', hex_mask='7')
    time.sleep(0.05)
    write_pci('000003f0', hex_mask='7')
    time.sleep(0.05)


def flush_dma(dma='dma0'):
    log.info('Flushing DMA Pipeline')
    write_pci('03f0', hex_mask='CF0')
    cmd = ['pci', '-r', dma, '-o', '/dev/null', '--multipacket']
    safe_call(cmd)


def startup_sequence():
    NUMBER = 4
    yield NUMBER

    start_dma()
    write_pci('0x20001000', '0x9100')
    write_pci('0x05', '0x9040')
    yield True

    write_pci('0x3C1', '0x9040')
    logging.info("9040:  " + str(read_pci(1, '0x9040')[0]))
    yield True

    write_pci('0x3F1', '0x9040')
    logging.info("9040:  " + str(read_pci(1, '0x9040')[0]))
    yield True

    write_pci('0x3F0', '0x9040')
    logging.info("9040:  " + str(read_pci(1, '0x9040')[0]))
    yield True


def calibration_sequence():
    NUMBER = 15
    yield NUMBER

    # Removing Board Reset
    write_pci('0x2000003f0', '0x9040')
    yield True

    # SPI Fanout Programming...
    write_pci('0x083', '0x9068')
    yield True

    write_pci('0x6a2', '0x9068')
    yield True

    # PLL calibration start...
    # PLL Reset...
    write_pci('0x80000000', '0x9060')
    yield True

    # Set CH_0 (FPGA) clock FPGA ...
    write_pci('0x00050000', '0x9060')
    yield True

    # Set CH_3 clock fanout ...
    write_pci('0x00050003', '0x9060')
    yield True

    # Set CH_4 clock ADC 1 ...
    write_pci('0x00050004', '0x9060')
    yield True

    # Set CH_5 clock ADC 2 ...
    write_pci('0x00050005', '0x9060')
    yield True

    # Set CH_6 clock ADC 3 ...
    write_pci('0x00050006', '0x9060')
    yield True

    # Set CH_7 clock ADC 4 ...
    write_pci('0x00050007', '0x9060')
    yield True

    # Set R8 ...
    write_pci('0x10000908', '0x9060')
    yield True

    # Set R11 ...
    write_pci('0x0082800B', '0x9060')
    yield True

    # Set R13 ...
    write_pci('0x029F400D', '0x9060')
    yield True

    # Set R14 (F_out and Global_EN => ON) ...
    write_pci('0x0830040E', '0x9060')
    yield True

    # Set R15 ...
    write_pci('0xD000100F', '0x9060')
    yield True


def synchronisation_sequence():
    NUMBER = 2
    yield NUMBER

    # Send the PLL sync signals ...
    write_pci('0x1003f0', '0x9040')
    yield True

    write_pci('0x0003f0', '0x9040')
    yield True

def write_value_sequence():
    NUMBER = 8
    yield NUMBER

    # Set_FPGA_clock_delay.sh 0
    config.set_fpga_delay(config.get('fpga_delay'))
    yield True

    # Set_Delay_chip.sh 16 16 16 16
    factors = [config.get('chip_1_delay'), config.get('chip_2_delay'),
               config.get('chip_3_delay'), config.get('chip_4_delay')]
    config.set_chip_delay([0, 1, 2, 3], factors)
    yield True

    # Set_TH_Delay.sh 12
    config.set_th_delay(config.get('th_delay'))
    yield True

    # Set_ADC_1_Delay.sh 4
    config.set_adc_delay(0, config.get('adc_1_delay'))
    yield True

    # Set_ADC_2_Delay.sh 4
    config.set_adc_delay(1, config.get('adc_2_delay'))
    yield True

    # Set_ADC_3_Delay.sh 4
    config.set_adc_delay(2, config.get('adc_3_delay'))
    yield True

    # Set_ADC_4_Delay.sh 4
    config.set_adc_delay(3, config.get('adc_4_delay'))
    yield True

    write_pci('{0:08x}'.format(config.get('orbits_observe')), '0x9020')
    yield True

    write_pci('{0:08x}'.format(config.get('orbits_skip')), '0x9028')
    yield True


def stop_board():
    write_pci('0x01', '0x9040')
    stop_dma()


def soft_reset():
    write_pci('0x1', '0x9040', hex_mask='0x1')
    time.sleep(1)
    write_pci('0x0', '0x9040', hex_mask='0x1')

def is_conneced():
    try:
        read_pci(1, '0x9040')
        return True
    except BoardError:
        return False
    except InterfaceNotFoundError:
        return None


def is_active():
    control = read_pci(1, '0x9040')[0]
    control_bits = '{0:032b}'.format(int(control, 16))
    return control_bits[22:26] == "1111"


def start_pilot_bunch_emulator():
    log.info('Start pilot bunch emulator')
    write_pci('400003f0', hex_mask='400003F0')
    time.sleep(0.005)
    write_pci('03f0', hex_mask='CF0')


def start_acquisition():
    log.info('Start acquisition')
    write_pci('1', '4', hex_mask='1')
    time.sleep(0.005)
    write_pci('00bf0', hex_mask='CF0')


def stop_acquisition():
    log.info('Stop acquisition')
    write_pci('003f0', hex_mask='CF0')
    time.sleep(0.005)


def enable_transfer():
    log.info('Enable data transfer')
    write_pci('007f0', hex_mask='CF0')
    time.sleep(0.005)


def write_data(filename, dma='dma0'):
    log.info('Write data to {}'.format(filename))
    cmd = ['pci', '-r', dma, '-o', filename, '--multipacket']
    safe_call(cmd)
    write_pci('003f0', hex_mask='CF0')


def run_status(script_path):
    cmd = ['bash', script_path]
    output = safe_call(cmd)
    log.info(output)


def acquire_data(filename, simulate=False):
    if simulate:
        start_pilot_bunch_emulator()

    start_acquisition()
    wait_for_revolutions()
    stop_acquisition()
    enable_transfer()
    write_data(filename)
    flush_dma()


def wait_for_revolutions():
    n = read_pci(1, '0x9020', decimal=True)[0]  # Get the amount of orbits to observe
    # n = 1 # Use this for debugging purposes if no board is connected
    spin_time_ns = kcg_config.tRev * n
    time.sleep(spin_time_ns * 1.1)  # 10% Safety margin


class StatusStorage(object):
    """
    Class used as Container for various storage purposes
    """
    def __getattr__(self, item):
        try:
            # return self.old_getattr(item)
            return object.__getattribute__(self, item)
        except AttributeError:
            print item+" was not set, None is returned"
            return None



status = StatusStorage()  # make status globally acessible

