"""
This is the interface to the backend.
It is used to make the backend easily interchangable.
All Functions that interface directly with the backend are prefixed with be_
Functions only used internal in this module will be prefixed _bif_
"""

import logging
import time
from datetime import datetime as dt
import os
from PyQt4 import QtGui, QtCore
import numpy as np
from backend import board
from backend import io
from backend import dataset
from groupedelements import Buttons, Elements, live_plot_windows
import storage
from .. import config
import kcgwidget as kcgw
from log import log

tr = kcgw.tr

livePlotData = None


def initStatus(st):  # TODO: Better solution? or implement all necesseary status things
    """
    Initialize Status variables. These variables are used to transfer status variables over different modules.
    :param st: variable to use (most likely a DummyStorage instance)
    :return: -
    """
    st.continuous_read = False
    st.calibrated = False
    st.synced = False
    st.defaults_set = False
    st.status_text = tr("sw", "Ready")
    st.time_scan = False
    st.wait = False
    st.last_file = None
    st.board_connected = True
    st.continuous_interval = 1000


initStatus(board.status)


# -----------[ Backend Interface ]----------------------
def _bif_enable_wait_cursor():
    """
    Show the "Wait Cursor"
    """
    QtGui.QApplication.setOverrideCursor(QtGui.QCursor(QtCore.Qt.WaitCursor))


def _bif_disable_wait_cursor():
    """
    Show the normal Cursor
    """
    QtGui.QApplication.restoreOverrideCursor()


def _bif_continuous_read_is_enabled(popup_title_text=None):
    """
    Checks if continuous read is enabled and if yes shows a popup dialog to ask if it shall be disabled
    :popup_title_text: Text to display in the popup that asks to disable continuous read
    :return: bool (True if continuous read is enabled and not disabled in popup else False)
    """
    if board.status.continuous_read:
        popup = PopupDialog(tr("Dialog", "Continuous read is currently active!\nStop continuous read and proceed?"),
                            title=popup_title_text)
        popup.exec_()
        popup.deleteLater()
        if popup.get_return_value() is False:
            return False
        else:
            _bif_set_continuous_read_inactive()
            Elements.setChecked("continuous_read", False)
            return True
    else:
        return False


def bk_status_readout():
    """
    Read Status from board and update variables
    as well as enable and disable corresponding Buttons
    :return: -
    """
    # TODO: NOTE: Problem with this is that certain buttons will be greyed out until other buttons are pressed
    # TODO: even if only the gui is restarted and the board is in the same state
    if kcgw.testing:
        return  # TODO: IMPORTANT REMOVE THIS !!!!!
    if board.is_active():
        Buttons.setEnabled("after_start", True)
        Buttons.setEnabled("start_board", False)
        Buttons.setEnabled("continuous_read", True)

        # MenuItems.setEnabled("continuous_read", True)
    else:
        board.status.calibrated = False
        board.status.synced = False
        board.status.defaults_set = False
        Buttons.setEnabled("after_start", False)
        Buttons.setEnabled("continuous_read", False)
        # MenuItems.setEnabled("continuous_read", False)
    Buttons.setEnabled("synchronize", board.status.calibrated)
    Buttons.setEnabled("set_defaults", board.status.synced)
    Buttons.setEnabled("acquire", board.status.synced)
    Buttons.setEnabled("acquireTrigger", board.status.synced)
    # Elements.setEnabled("timing", board.status.defaults_set)
    Elements.setEnabled("timing", board.status.synced)

    storage.update_LED()


backup_readout = bk_status_readout


class PopupDialog(QtGui.QDialog):
    """
    Simple Class to show a popup dialog.
    """

    def __init__(self, text, title=None, parent=None):
        QtGui.QDialog.__init__(self, parent)
        self.text = text
        self.setWindowTitle(title if title else tr("Dialog", "User action required"))
        self.do_layout()
        self.return_value = False

    def do_layout(self):
        size = QtCore.QSize(200, 200)
        # self.setMaximumSize(size)
        self.setMinimumSize(size)
        box = QtGui.QVBoxLayout()
        self.text_label = QtGui.QLabel(self.text)
        self.text_label.setAlignment(QtCore.Qt.AlignCenter)
        box.addWidget(self.text_label)
        self.okay_btn = QtGui.QPushButton(tr("Button", "Ok"))
        self.okay_btn.setStyleSheet("padding: 15px;")
        self.okay_btn.clicked.connect(self.on_okay)
        box.addWidget(self.okay_btn)
        box.addSpacerItem(QtGui.QSpacerItem(1, 20))
        self.cancel_btn = QtGui.QPushButton(tr("Button", "Cancel"))
        self.cancel_btn.setStyleSheet("padding: 15px;")
        self.cancel_btn.clicked.connect(self.on_cancel)
        box.addWidget(self.cancel_btn)
        self.setLayout(box)

    def on_okay(self):
        self.return_value = True
        self.close()

    def on_cancel(self):
        self.close()

    def get_return_value(self):
        return self.return_value


def bk_start_board():
    """
    Start the Board.
    This will set initial Registers to power up the Board.
    :return: -
    """
    _bif_enable_wait_cursor()
    log(additional="Starting Board - following values are probably default values")
    sequence = board.startup_sequence()
    number = sequence.next()

    if _bif_continuous_read_is_enabled(tr("Button", "Start Board")):
        _bif_disable_wait_cursor()
        return

    try:
        if board.is_active():
            _bif_disable_wait_cursor()
            bk_status_readout()
            return

        logging.info("##### Activating Board #####")
        sequence.next()
        time.sleep(1.0)
        dialog1 = PopupDialog(tr("Button", "Switch On the power supply --> FIRST <--"))
        dialog1.exec_()
        dialog1.deleteLater()

        if not dialog1.get_return_value():
            logging.error("Starting procedure canceled")
            _bif_disable_wait_cursor()
            return

        logging.info("Switch ON T/Hs")
        sequence.next()
        time.sleep(0.1)
        dialog2 = PopupDialog(tr("Dialog", "Switch On the power supply --> SECOND <--"))
        dialog2.exec_()
        dialog2.deleteLater()

        if not dialog2.get_return_value():
            logging.info("Starting procedure canceled")
            _bif_disable_wait_cursor()
            return

        logging.info("Switch ON ADCs")
        sequence.next()
        time.sleep(0.1)

        sequence.next()
        time.sleep(1.0)

        for step in sequence:
            time.sleep(0.1)

        logging.info("Board started successfully!")
    except board.BoardError as e:
        logging.error("Starting board failed: {}".format(str(e)))

    bk_update_config('header', bk_get_config('header'))
    _bif_disable_wait_cursor()
    bk_status_readout()


class _bif_ProgressBar(QtGui.QProgressBar):
    """
    Simple Progressbar class.
    """

    def __init__(self, min, max, text):
        super(_bif_ProgressBar, self).__init__()
        self.setRange(min, max)
        self.setMaximumHeight(18)
        kcgw.statusbar.clearMessage()
        self.label = QtGui.QLabel(text)
        kcgw.statusbar.insertWidget(0, self.label)
        kcgw.statusbar.insertWidget(1, self)
        self.setValue(0)
        QtGui.qApp.processEvents()

    def remove(self, timeout=None):
        """
        Remove this instance of a progressbar
        :param timeout: the time from calling this function to wanishing of the progressbar
        :return: -
        """

        def remove_progressbar():
            kcgw.statusbar.removeWidget(self)
            kcgw.statusbar.removeWidget(self.label)
            kcgw.statusbar.showMessage(board.status.status_text)
            self.destroy()

        if timeout:
            QtCore.QTimer.singleShot(timeout, remove_progressbar)
        else:
            remove_progressbar()


thread = None
cal = None


def bk_calibrate(do_the_rest=None):
    """
    Send commands to the board that will enable it to calibrate itself.
    This function checks if a read command is still running. BUT: It does not check if
    the board is acquiring or something like this.
    So Another instance of KCG can still be acquiring or calibrating at the same time. This can be dangerous.
    :param do_the_rest: function to call after calibration. This is used when "Prepare Board" is pressed.
    :return: -
    """
    log(additional="Calibrate")
    global thread
    global cal
    if thread is not None:
        logging.info("Calibration already running")
        return
    _bif_enable_wait_cursor()
    if _bif_continuous_read_is_enabled(tr("Button", "Calibrate Board")):
        _bif_disable_wait_cursor()
        return

    sequence = board.calibration_sequence()
    number = sequence.next()
    progressbar = _bif_ProgressBar(0, number, tr("sw", "Calibrating"))

    class Calibrate(QtCore.QObject):
        """
        Class to use as thread class. NOTE: this is not used at the moment.
        """
        update_progressbar_signal = QtCore.pyqtSignal(int)
        finished = QtCore.pyqtSignal()

        def calibrate(self):
            try:
                logging.info('Started Board Calibration')
                for idx, step in enumerate(sequence):
                    time.sleep(0.5)
                    self.update_progressbar_signal.emit(idx)
                board.status.calibrated = True
            except board.BoardError as e:
                logging.error("Calibration failed: {}".format(str(e)))
                _bif_disable_wait_cursor()
                # self.do_status_readout()
                bk_status_readout()
                self.finished.emit()
                return

            logging.info("Board Calibration successful!")
            bk_status_readout()
            self.finished.emit()

    def thread_quit():
        global thread
        progressbar.remove(0)
        # thread.quit()
        # thread.wait()
        thread = None
        _bif_disable_wait_cursor()
        if do_the_rest:  # execute sync and set defaults (this is set if prepare board was pressed)
            do_the_rest()

    # thread = QtCore.QThread()
    cal = Calibrate()
    cal.update_progressbar_signal.connect(progressbar.setValue)
    cal.finished.connect(thread_quit)
    # cal.moveToThread(thread)
    # thread.started.connect(cal.calibrate)
    # thread.start()
    cal.calibrate()


def bk_sync_board():
    """
    Sends commands to the board to sync with triggers.
    :return: -
    """
    _bif_enable_wait_cursor()
    log(additional="Synchronize")
    if _bif_continuous_read_is_enabled(tr("Dialog", "Synchronize Board")):
        _bif_disable_wait_cursor()
        return

    try:
        sequence = board.synchronisation_sequence()
        sequence.next()  # skip number
        progressbar = _bif_ProgressBar(0, 100, tr("sw", "Synchronizing"))
        logging.info("Synchronize PLLs")
        sequence.next()
        for i in range(1, 101):
            time.sleep(0.01)
            progressbar.setValue(i)
        sequence.next()
        progressbar.remove(0)
    except board.BoardError as e:
        logging.error("Synchronization failed: {}".format(str(e)))
        progressbar.remove(0)
        _bif_disable_wait_cursor()
        bk_status_readout()
        return

    logging.info("Board synchronization successful!")
    # self.set_defaults_button.setEnabled(True)
    board.status.synced = True
    _bif_disable_wait_cursor()
    bk_status_readout()


def bk_write_values(defaults=False):
    """
    Write values to board.
    :param defaults: (bool) if True Writes default values
    :return: -
    """
    _bif_enable_wait_cursor()
    if defaults:
        log(additional="Set Default Values")
    else:
        log(additional="Update Values on board")
    if _bif_continuous_read_is_enabled(tr("Dialog", "Update Values on Board")):
        _bif_disable_wait_cursor()
        return

    try:
        sequence = board.write_value_sequence()
        number = sequence.next()
        if defaults:
            board.config._set_defaults()
            progressbar = _bif_ProgressBar(0, number, tr("sw", "Setting Defaults"))
            logging.info("##### Setting default Values #####")
        else:
            progressbar = _bif_ProgressBar(0, number, tr("sw", "Updating Values on Board"))
            logging.info("##### Update Values #####")
        for idx, step in enumerate(sequence):
            time.sleep(0.1)
            progressbar.setValue(idx)

        board.status.defaults_set = True
        progressbar.remove(0)

    except board.BoardError as e:
        logging.error("Updating Values failed: {}".format(str(e)))
        progressbar.remove(0)
        _bif_disable_wait_cursor()
        bk_status_readout()
        return

    if defaults:
        logging.info("Default values set successfully!")
    else:
        logging.info("Updated values successfully!")
    _bif_disable_wait_cursor()
    bk_status_readout()
    board.config.notify_all_observers()


def bk_stop_board():
    """
    Stops the board and shuts it down
    :return: -
    """
    _bif_enable_wait_cursor()
    log(additional="Stop Board")
    if _bif_continuous_read_is_enabled(tr("Dialog", "Soft Reset Board")):
        _bif_disable_wait_cursor()
        return

    try:
        logging.info("##### Switching Off Board #####")
        board.stop_board()
        time.sleep(0.5)
    except board.BoardError as e:
        logging.error("Sequence failed: {}".format(str(e)))
        _bif_disable_wait_cursor()
        bk_status_readout()
        return

    logging.info("Board switched off successfully!")
    Buttons.setEnabled("after_start", False)
    Buttons.setEnabled("start_board", True)
    board.status.calibrated = False
    _bif_disable_wait_cursor()
    bk_status_readout()


def bk_soft_reset():
    """
    Perform a soft reset.
    :return:
    """
    _bif_enable_wait_cursor()
    log(additional="Soft Reset")
    if _bif_continuous_read_is_enabled(tr("Dialog", "Soft Reset Board")):
        _bif_disable_wait_cursor()
        return

    try:
        logging.info("Soft-Resetting Board ...")
        board.soft_reset()
    except board.BoardError as e:
        logging.error("Sequence failed: {}".format(str(e)))
        _bif_disable_wait_cursor()
        bk_status_readout()
        return

    logging.info("Soft-Reset successful.")
    _bif_disable_wait_cursor()
    bk_status_readout()
    board.config.update('header', True)  # Reset Header if Header bit is erased after soft reset


def bk_update_config(key, value, silent=False):
    """
    Interface to the update command of the BoardConfiguration class.
    :param key: Key to update
    :param value: Value to set for key
    :return: -
    """
    try:
        if silent:
            board.config.updateSilent(key, value)
        else:
            board.config.update(key, value)
    except board.BoardError as e:
        logging.error("Setting value of {} failed: {}".format(key, str(e)))


def bk_get_config(key):
    """
    Interface to the get command of the BoardConfiguration class.
    :param key: Key to get the value for
    :return: value stored for key
    """
    return board.config.get(key)

def bk_get_status():
    """
    Interface to the get_status of the board
    :return: status dictionary
    """
    return board.get_status()

def bk_change_num_of_orbits(value, silent=False):
    """
    Send new number of orbits to board and update in config
    :param value: the value to send
    :param silent: (bool) if True do not inform observers on update
    :return: -
    """
    bk_update_config("orbits_observe", value, silent=silent)


def bk_change_num_of_skipped_orbits(value, silent=False):
    """
    Send new number of orbits to skip to board and update in config
    :param value: the value to send
    :param silent: (bool) if True do not inform observers on update
    :return: -
    """
    bk_update_config("orbits_skip", value, silent=silent)


def bk_change_count(value, silent=False):
    """
    Change the number of acquisitions you want to make.
    :param value: (int) Number of acquisitions
    :param silent: (bool) if True do not inform observers on update
    :return: -
    """
    bk_update_config("acquisition_count", value, silent=silent)


def bk_change_wait(value, silent=False):
    """
    Change the time between acquisitions.
    :param value: (bool) Time in seconds
    :param silent: (bool) if True do not inform observers on update
    :return: -
    """
    bk_update_config("orbits_wait_time", value, silent=silent)


def bk_change_build_spectrograms(value, silent=False):
    """
    Change if spectrograms are built or not)
    :param value: (bool) True or False built or not
    :param silent: (bool) if True do not inform observers on update
    :return:
    """
    bk_update_config("build_spectrograms", value, silent=silent)


def bk_change_pilot_bunch(value, silent=False):
    """
    Change if pilot bunch is simulated
    :param value: (bool) True or False to simulate or not
    :param silent: (bool) if True do not inform observers on update
    :return:
    """
    bk_update_config("pilot_bunch", value, silent=silent)


def _bif_iterate_spectrograms(path):
    """
    Built Spectrograms line by line
    :param path: where to built the spectrogram
    :return: -
    """
    if not os.path.isdir(str(path)):
        return

    transform = dataset.data.fft(1, frm=0, to=-1)
    for i in range(config.bunches_per_turn - 1):
        filename = os.path.join(storage.storage.save_location, storage.storage.subdirname, str(path), "%i.hsp" % i)
        write_header = False
        if not os.path.isfile(filename):
            write_header = True
        f = open(filename, 'ab')
        if write_header:
            f.write("#hsp\n")  # heb spectrogram magic number
            # f.write(b'\x40')  # 64bit
            # f.write('64')  # 64bit
            # f.write(sytruct.pack('>I', self.number_of_skipped_orbits_spinbox.value()))
            f.write("#" + str(board.config.get("orbits_skip")))
            f.write("\n")
        line = transform[i, :]
        # f.write(struct.pack('>I', len(line)))
        # f.write(str(len(line)))
        f.write('{:0.3f} '.format(time.time()))
        # data = pArr('d', line)
        # data.tofile(f)
        for e in line:
            f.write("%s " % np.absolute(e))
        f.write("\n")
        f.close()


def _bif_read_data_and_save():
    """
    Tell the pci command to start acquisition and save data
    Also generates the filename from settings
    :return:
    """
    now = time.time()
    if not os.path.isdir(str(storage.storage.save_location + '/' + storage.storage.subdirname)):
        os.makedirs(str(storage.storage.save_location + '/' + storage.storage.subdirname))
    filename = storage.storage.save_location + '/' + storage.storage.subdirname + '/{:0.3f}.out'.format(now)
    board.status.last_file = filename

    try:
        simulate = board.config.get("pilot_bunch")
        board.acquire_data(filename, simulate=simulate)
        _bif_read_and_update_data_from_file(filename)
    except board.BoardError as e:
        logging.error("Reading failed: {}".format(str(e)))


def _bif_read_and_update_data_from_file(filename):
    """
    Proxy function for _bif_read_and_update to call with correct read_func
    :param filename: filename to read data from
    :return: -
    """
    _bif_read_and_update(io.read_from_file, str(filename))


def _bif_read_and_update_data_from_string(raw_data):
    """
    Proxy function for _bif_read_and_update to call with correct read_func
    :param raw_data: Data as string
    :return: -
    """
    _bif_read_and_update(io.read_from_string, raw_data)


def _bif_read_and_update(read_func, *args):
    """
    Function to read data from file or string (depending on read_func) and update plots etc.
    :param read_func: function to use to read data
    :param args: filename or raw_data (see _bif_read_and_update_from_{filename, string}
    :return: -
    """
    _bif_enable_wait_cursor()

    header = board.config.get('header')
    # TODO: force_read: meaning ignore cache and read new -> in the old gui this was a commandline option
    # TODO: cache_data: meaning cache already processed numpy data -> in the old gui this was a commandline option
    if live_plot_windows.hasWindows():
        data = read_func(*args, force=False, header=header, cache=False)

        if not io.is_data_consistent(data):
            kcgw.consistency_updater(False)
            # self.data_consistency_led.set_color((0, 190, 0))
            kcgw.statusbar.showMessage(tr("Dialog", "Data is inconsistent!"))
            if read_func == io.read_from_string:
                logging.info("Data is inconsistent")
            else:
                logging.info("Data is inconsistent - file: " + args[0])
        else:
            kcgw.consistency_updater(True)

        for plotwin in live_plot_windows.getWindows():
            plotwin.plot_live(data=data)
    else:
        kcgw.consistency_updater(None)

    _bif_disable_wait_cursor()


def bk_acquire():
    """
    Toggle Acqisition
    :return:
    """
    if not bk_get_config('use_trigger'):
        if board.status.acquisition == True:
            log(additional="Manually Stopped Acquisition\nPerformed Acquisitions: " + str(
                storage.storage.current_acquisition))
            _bif_stop_acquisition()
        else:
            _bif_start_acquisition()
    else:
        bk_toggle_wait_on_trigger()

def _bif_stop_acquisition():
    """
    Stop acquisition
    This does stop the timer started by _bif_start_acquisition()
    :return: -
    """
    board.status.acquisition = False
    board.acquisition_progressbar.remove(0)
    board.acquisition_timer.stop()
    for elem in Elements.getElements("acquireTrigger"):
        if isinstance(elem, QtGui.QShortcut):
            continue
        elem.setIcon(QtGui.QIcon(config.install_path + config.startIcon))
        elem.setText(tr("Button", "Start Acquisition"))
    # self.continuous_read_checkbox.setEnabled(True)
    # self.acquisition_progressbar.reset()
    Elements.setEnabled('acquire', True)


def _bif_start_acquisition():
    """
    Start acquisition.
    This will start a timer to automatically acquire data.
    :return: -
    """
    log(additional="Started Acquisition")
    board.status.acquisition = True
    Elements.setEnabled('acquire', False, exclude=Elements.getElements('acquireTrigger'))
    for elem in Elements.getElements("acquireTrigger"):
        if isinstance(elem, QtGui.QShortcut):
            continue
        elem.setIcon(QtGui.QIcon(config.install_path + config.stopIcon))
        elem.setText(tr("Button", "Stop Acquisition"))
    storage.storage.acquisition_timer = QtCore.QTimer()
    num_acquisitions = board.config.get("acquisition_count")
    acquisition_progressbar = _bif_ProgressBar(0, num_acquisitions, tr("sw", "Acquiring"))
    # storage.storage.acquisition_progressbar = acquisition_progressbar

    # We increase already once because we do a single acquisition before the
    # timer is started, otherwise we have to wait until the timer fires the
    # first time.
    storage.storage.current_acquisition = 1
    _bif_read_data_and_save()

    if board.config.get("build_spectrograms"):
        spectrogram_dir = storage.storage.save_location + '/' + storage.storage.subdirname + "/spectrograms_{:0.3f}".format(
            time.time())
        os.makedirs(spectrogram_dir)
        _bif_iterate_spectrograms(spectrogram_dir)  # TODO: not here?

    acquisition_progressbar.setValue(storage.storage.current_acquisition)

    def on_timeout():
        if storage.storage.current_acquisition < num_acquisitions:
            storage.storage.current_acquisition += 1
            acquisition_progressbar.setValue(storage.storage.current_acquisition)
            _bif_read_data_and_save()
            if board.config.get("build_spectrograms"):
                _bif_iterate_spectrograms(spectrogram_dir)  # TODO: not here ?
        else:
            log(additional="Stopped Acquisition")
            _bif_stop_acquisition()
            acquisition_progressbar.remove(0)

    storage.storage.acquisition_timer.timeout.connect(on_timeout)
    storage.storage.acquisition_timer.start(board.config.get('orbits_wait_time') * 1000)


def bk_single_read():
    """
    Perform a single read of data
    :return:
    """
    Elements.setEnabled("acquire", False)
    _bif_read_data_and_save()
    log(additional="Single Read\nFilename: " + board.status.last_file.split('/')[-1])
    Elements.setEnabled("acquire", True)


def _bif_set_continuous_read_active():
    """
    Enable continuous read
    :return: -
    """
    Elements.setEnabled("acquire", False)
    Elements.setEnabled("continuous_read", True)
    board.status.continuous_read = True


def _bif_set_continuous_read_inactive():
    """
    Disable continuous read
    :return: -
    """
    if board.status.continuous_read:
        board.status.continuous_read = False
        board.continuous_read_timer.stop()
    Elements.setEnabled('acquire', True)


def bk_continuous_read(interval=None):
    """
    Toggle continuous read
    :param interval: Time between two consecutive reads.
    :return: -
    """
    if not board.status.continuous_read:
        _bif_set_continuous_read_active()
        _bif_continuous_read(interval)
    else:
        _bif_set_continuous_read_inactive()


def _bif_continuous_read(interval=None):
    """
    Perform continuous read based on a timer.
    :param interval:
    :return:
    """
    if interval is not None:
        storage.storage.continuous_interval = interval
    board.continuous_read_timer = QtCore.QTimer()
    logging.info("Start continuous read")

    def continuous_read_step():
        if board.status.continuous_read:
            _bif_read_data()
            # board.continuous_read_timer.start(interval)
            board.continuous_read_timer.singleShot(storage.storage.continuous_interval, continuous_read_step)
        #        else:
        #            board.continuous_read_timer.stop()


        #    board.continuous_read_timer.timeout.connect(continuous_read_step)
        #    board.continuous_read_timer.start(storage.storage.continuous_interval)

    board.continuous_read_timer.singleShot(storage.storage.continuous_interval, continuous_read_step)


def _bif_read_data():
    """
    Reads data acquired by board.
    :return:
    """
    try:
        if board.config.get('pilot_bunch'):
            board.start_pilot_bunch_emulator()

        board.start_acquisition()
        board.wait_for_revolutions()
        board.stop_acquisition()
        board.enable_transfer()
        cmd = ['pci', '-r', 'dma0', '-o', '/dev/stdout', '--multipacket']
        data_raw = board.safe_call(cmd)

        # board.flush_dma()  This takes long and does not do any good :)
        # Sadly, the PCI software also outputs the Information, that it has
        # written data to the desired file.  This is recorded by our
        # software and needs to be removed, so we split the returned string
        # at the 'Writting data to file blah, blah" Btw: YES this
        # 'Writting' is actually correct. No, i don't know why it is
        # grammatically wrong in the first place :D
        data_split = data_raw.split('Writting')[0]
        _bif_read_and_update_data_from_string(data_split)
    except board.BoardError as e:
        logging.error("Reading failed: {}".format(str(e)))


def bk_board_connected():
    """
    Interface to the board to check if it is connected.
    :return: -
    """
    return board.is_conneced()


def bk_get_temperature():
    """
    Get Temperature from board and format it
    :return: -
    """
    fpga_temp_raw_hex = board.read_pci(1, '0x9110')[0]
    fpga_temp_raw_hex = fpga_temp_raw_hex[-3:]
    fpga_temp_raw_bin = '{0:012b}'.format(int(fpga_temp_raw_hex, 16))
    fpga_temp_encoded = board.get_dec_from_bits(fpga_temp_raw_bin, 9, 0)
    fpga_temp_celsius = '{0:2.2f}'.format(((fpga_temp_encoded * 503.975) / 1024) - 273.15)
    return fpga_temp_celsius


backup_get_temp = bk_get_temperature


def bk_time_scan(c_frm, c_to, f_frm, f_to, ts_pbar, plot_func, orbits_observe=None, orbits_skip=None):
    """
    Toggle Timescan.
    :param c_frm: (int) From value for Coarse scan
    :param c_to:  (int) To value for Coarse scan
    :param f_frm: (int) From value for Fine scan
    :param f_to: (int) To value for fine scan
    :param ts_pbar: Handle to the Timescan Progressbar
    :param plot_func: Function to plot when timescan ended.
    :param orbits_observe: Number of orbits to observe for the timescan (original values will be restored after timescan)
    :param orbits_skip: Number of orbits to skipfor the timescan (original values will be restored after timescan)
    :return: -
    """
    if board.status.time_scan:
        _bif_stop_time_scan(ts_pbar)
    else:
        _bif_start_time_scan(c_frm, c_to, f_frm, f_to, ts_pbar, plot_func, orbits_observe, orbits_skip)


def _bif_stop_time_scan(ts_pbar):
    """
    Stop the timescan. This stops the timer.
    :param ts_pbar: Timescan Progressbar handle
    :return: -
    """
    Elements.getElements("start_time_scan")[0].setText(tr("Button", "Start time scan"))
    board.status.time_scan = False
    ts_pbar.reset()


tst = None  # Ugly but necessary for the thread not to be killed when function ends (which is almost immediately after start)
thread_ts = None


def _bif_start_time_scan(c_frm, c_to, f_frm, f_to, timescan_progressbar, plot_func, orbits_observe, orbits_skip):
    """
    Start the timscan. This starts the timer
    :param c_frm: From value for coarse scan
    :param c_to: To value for coarse scan
    :param f_frm: From value for fine scan
    :param f_to: To value for fine scan
    :param timescan_progressbar: Handle for the timescanprogressbar
    :param plot_func: Function to use to plot the data
    :param orbits_observe: Number of orbits to observe for the timescan (original values will be restored after timescan)
    :param orbits_skip: Number of orbits to skipfor the timescan (original values will be restored after timescan)
    :return: -
    """
    if thread_ts is not None:
        logging.info("Time scan already running")
        return
    board.status.time_scan = True
    Elements.getElements("start_time_scan")[0].setText(tr("Button", "Stop time scan"))
    if c_frm > c_to:
        logging.info('Coarse Scan Interval is invalid: (%i > %i)' % (c_frm, c_to))
        return
    if f_frm > f_to:
        logging.info('Fine Scan Interval is invalid: (%i > %i)' % (f_frm, f_to))
        return

    storage.th_old = board.config.get('th_delay')
    storage.chip_1_old = board.config.get('chip_1_delay')
    storage.chip_2_old = board.config.get('chip_2_delay')
    storage.chip_3_old = board.config.get('chip_3_delay')
    storage.chip_4_old = board.config.get('chip_4_delay')

    minimum = [None, None, None, None]
    maximum = np.zeros((4, 3))
    heatmap = np.zeros((4, (f_to - f_frm + 1), (c_to - c_frm + 1)))
    timescan_progressbar.setRange(1, ((f_to - f_frm) + 1) * ((c_to - c_frm) + 1))

    class thread_time_scan(QtCore.QObject):
        pbarSignal = QtCore.pyqtSignal(int)
        stopSignal = QtCore.pyqtSignal()
        finished = QtCore.pyqtSignal()

        def __init__(self, c_frm, c_to, f_frm, f_to, timescan_progressbar):
            super(thread_time_scan, self).__init__()
            self.c_frm = c_frm
            self.c_to = c_to
            self.f_frm = f_frm
            self.f_to = f_to
            self.timescan_progressbar = timescan_progressbar

        def time_scan(self):
            Elements.setEnabled('acquire', False, exclude=Elements.getElements('start_time_scan'))
            if orbits_observe:
                storage.orbits_observe_before_timescan = board.config.get(
                    "orbits_observe")  # save old values to restore after timescan
                board.config.update("orbits_observe", orbits_observe)
                bk_change_num_of_orbits(orbits_observe)
            if orbits_skip:
                storage.orbits_skip_before_timescan = board.config.get("orbits_skip")
                board.config.update("orbits_skip", orbits_skip)
                bk_change_num_of_skipped_orbits(orbits_skip)

            c_step = 0
            for coarse in range(self.c_frm, self.c_to + 1):
                try:
                    board.config.set_delay(coarse)
                except board.BoardError as e:
                    self.stopSignal.emit()
                    self.finished.emit()
                    return

                f_step = 0
                for fine in range(self.f_frm, self.f_to + 1):
                    board.config.set_chip_delay([0, 1, 2, 3], [fine, fine, fine, fine])

                    try:
                        if bk_get_config('pilot_bunch') is True:
                            board.start_pilot_bunch_emulator()

                        board.start_acquisition()
                        board.wait_for_revolutions()  # Wait before asking for data
                        board.stop_acquisition()
                        board.enable_transfer()

                        # -----------[ IMPORTANT ]---------------------
                        if not kcgw.testing:
                            cmd = ['pci', '-r', 'dma0', '-o', '/dev/stdout', '--multipacket']
                            data_raw = board.safe_call(cmd)
                            board.flush_dma()
                        # ----------------------------------------------
                        else:
                            f_name = "/home/blaxxun/Documents/Hiwi/KaptureSimulator/timescan/" + str(
                                coarse) + "_" + str(fine) + ".str"
                            f = open(f_name, 'r')
                            data_raw = f.read()

                        # The PCI software not only prints the desired data but also some additional information.
                        # This information has to be removed here.
                        # To do so we split the output string from PCI at "Writting" (Note: Writting is correct as
                        # this is a typo in the PCI driver)
                        data_split = data_raw.split('Writting')[0]
                        data = io.read_from_string(data_split, force=True, cache=False)
                    except board.BoardError as e:
                        # QtGui.QMessageBox.critical(self, "Board communication", str(e)) # TODO: remove this if moved to thread again
                        # time_scan_stop()
                        # _bif_stop_time_scan(self.timescan_progressbar)
                        self.stopSignal.emit()
                        self.finished.emit()
                        return

                    for adc in range(4):
                        buckets = data.array[:, adc:adc + 1].reshape(-1)
                        heatmap[adc, f_step, c_step] = float(buckets.sum()) / buckets.shape[0]
                        if heatmap[adc, f_step, c_step] > maximum[adc, 0]:
                            maximum[adc, 0] = heatmap[adc, f_step, c_step]
                            maximum[adc, 1] = coarse
                            maximum[adc, 2] = fine
                        if minimum[adc] is None or minimum[adc] > heatmap[adc, f_step, c_step]:
                            minimum[adc] = heatmap[adc, f_step, c_step]

                    self.pbarSignal.emit(((c_step * (f_to - f_frm + 1)) + f_step) + 1)

                    # GUI is blocked in our tight loop. Give it an opportunity to handle events
                    QtGui.QApplication.processEvents()  # remove this if moved to thread again
                    if board.status.time_scan is False:
                        # Time Scan Stop is already performed by button press. Nothing else to do but leave
                        self.finished.emit()
                        return
                    f_step += 1
                c_step += 1
            self.finished.emit()

    def finished(timescan_progressbar):
        global thread_ts
        thread_ts.quit()
        thread_ts.wait()
        thread_ts = None
        _bif_stop_time_scan(timescan_progressbar)
        Elements.setEnabled('acquire', True, exclude=Elements.getElements('start_time_scan'))
        if orbits_observe:
            board.config.update("orbits_observe", storage.orbits_observe_before_timescan)  # restore values
            bk_change_num_of_orbits(storage.orbits_observe_before_timescan)
        if orbits_skip:
            board.config.update("orbits_skip", storage.orbits_skip_before_timescan)
            bk_change_num_of_skipped_orbits(storage.orbits_skip_before_timescan)
        board.config.set_delay(storage.th_old)
        board.config.set_chip_delay([0, 1, 2, 3],
                                    [storage.chip_1_old, storage.chip_2_old, storage.chip_3_old, storage.chip_4_old])

        m = [np.min(heatmap[heatmap != 0]), np.max(heatmap)]  # this gives the same levels for all 4 adcs
        plot_func(heatmap, levels=m, newTitle=str(tr("sw", "Coarserange:{c_f}-{c_t} ; Finerange:{f_f}-{f_t}")).format(
            c_f=c_frm,
            c_t=c_to,
            f_f=f_frm,
            f_t=f_to),
                  maxima=maximum
                  )

        now = time.time()
        if not os.path.isdir(str(storage.storage.save_location + '/' + storage.storage.subdirname + '/timescan')):
            os.makedirs((storage.storage.save_location + '/' + storage.storage.subdirname + '/timescan'))
        filename = storage.storage.save_location + '/' + storage.storage.subdirname + '/timescan/timescan_{:0.3f}.out'.format(
            now)
        f = open(filename, 'wr')

        for adc in range(4):
            f.write("#ADC_%s\n" % adc)
            for coarse, curr_cor in enumerate(np.transpose(heatmap[adc])):
                for fine, value in enumerate(curr_cor):
                    f.write("%i;%i;%f\n" % ((coarse + c_frm), (fine + f_frm), value))
            f.write('\n')

        f.close()
        f = open(filename + '.gnuplot', 'wr')
        f.write('set datafile separator ";"\n')
        f.write('set multiplot layout 2,2\n')
        f.write('unset key\n')
        for i in range(4):
            f.write('set label 1 "ADC_%i" at graph 0.7,0.95 font ",8"\n' % (i + 1))
            f.write('plot "%s" every :::%i::%i using 3 with lines\n' % (filename, i, i))
        f.write('unset multiplot\n')
        f.close()

        return

    global tst
    global thread_ts
    thread_ts = QtCore.QThread()
    tst = thread_time_scan(c_frm, c_to, f_frm, f_to, timescan_progressbar)
    tst.pbarSignal.connect(timescan_progressbar.setValue)
    tst.finished.connect(lambda: finished(timescan_progressbar))
    tst.stopSignal.connect(lambda: _bif_stop_time_scan(timescan_progressbar))
    tst.moveToThread(thread_ts)
    thread_ts.started.connect(tst.time_scan)
    thread_ts.start()
    # tst.wait()


def bk_check_for_board():
    """
    Check if board is connected
    Also overrides the bk_status_readout function with a function that does nothing (suppresses read attempts that
    generate errors - if no board is connected, there is nothing to read from)
    Also overrides the bk_get_temperature function as of the same reasons
    :return: -
    """
    global bk_status_readout, bk_get_temperature
    board_status = bk_board_connected()
    if board_status:
        if not board.status.board_connected:
            bk_status_readout = backup_readout
            bk_get_temperature = backup_get_temp
        board.status.board_connected = True

    else:
        Elements.setEnabled('no_board', False)

        def do_nothing():
            pass

        def no_temp():
            return "-"

        bk_status_readout = do_nothing
        bk_get_temperature = no_temp
        board.status.board_connected = False
        if board_status == False:
            board.status.status_text = tr("sw", "Board not connected")
        elif board_status == None:
            board.status.status_text = tr("sw", "Software Interface not found")


twt = None
thread_wait = None


def bk_toggle_wait_on_trigger(num_of_acquisitions=None, skip=None, timeout=None, method=None):
    """
    Toggle waiting for trigger signal to acquire
    :param num_of_acquisitions: number of acquisitions to wait for
    :param skip: how much trigger signals to skip between acquisitions
    :param count_label: the label that indicates in the gui how much acquisitions were performed
    :param timeout: the timeout for the pci to wait for date
    :param method: wait method to use
            1 for wait in pci command and 2 for waiting until register is set that KAPTURE has read data
            NOTE: this also means that method 1 enables simultaneous read and write to dma and method 2 does write and
            read sequentially
    :return:
    """
    global thread_wait
    if thread_wait is not None:
        twt.quit()
        for elem in Elements.getElements("acquireTrigger"):
            if isinstance(elem, QtGui.QShortcut):
                continue
            elem.setText(tr("Button", "Stopping Acquisition"))
            elem.setEnabled(False)
        # thread_wait.quit()
        # thread_wait.wait()
        # thread_wait = None
        # thread_wait.terminate()
    else:
        log(additional="Start wait on trigger")
        for elem in Elements.getElements("acquireTrigger"):
            if isinstance(elem, QtGui.QShortcut):
                continue
            elem.setIcon(QtGui.QIcon(config.install_path + config.stopIcon))
            elem.setText(tr("Button", "Stop Acquisition"))
        _bif_start_wait_on_trigger(num_of_acquisitions, skip, timeout, method)


def _bif_start_wait_on_trigger(num_of_acquisitions=None, skip=None, timeout=None, method=None):
    """
    Start waiting on external acquisition trigger. This starts the timer
    :param num_of_acquisitions: number of acquisitions to do
    :param count_label: Handle for the countlabel
    :param method: wait method to use
            1 for wait in pci command and 2 for waiting until register is set that KAPTURE has read data
            NOTE: this also means that method 1 enables simultaneous read and write to dma and method 2 does write and
            read sequentially
    :return: -
    """
    # TODO: docstring
    if thread_wait is not None:
        logging.info("Wait already running")
        return
    board.status.wait = True
    if not os.path.isdir(str(storage.storage.save_location + '/' + storage.storage.subdirname)):
        os.makedirs(str(storage.storage.save_location + '/' + storage.storage.subdirname))

    if not num_of_acquisitions:
        num_of_acquisitions = bk_get_config('acquisition_count')
    if not skip:
        skip = bk_get_config('trigger_skip')
    if not timeout:
        timeout = bk_get_config('trigger_timeout')
    if not method:
        method = bk_get_config('trigger_method')

    trigger_progressbar = _bif_ProgressBar(0, num_of_acquisitions, tr("sw", "Acquiring"))
    board.write_pci(hex(num_of_acquisitions), "9024")
    time.sleep(0.1)
    # board.safe_call(['pci', '-w', '902C', str(skip)])
    board.write_pci(hex(skip), "902C")
    time.sleep(0.1)

    class thread_wait_on_signal(QtCore.QObject):
        countUpdate = QtCore.pyqtSignal(int)
        stopSignal = QtCore.pyqtSignal()
        finished = QtCore.pyqtSignal()
        liveplot = QtCore.pyqtSignal(str)

        def __init__(self, num_of_acquisitions, path, timeout):
            super(thread_wait_on_signal, self).__init__()
            self.noa = num_of_acquisitions
            self.path = path
            self.timeout = timeout
            self._quit = False

            Elements.setEnabled('acquire', False, exclude=Elements.getElements('acquireTrigger'))

        def wait_rw_simul(self):
            board.write_pci('ff0', hex_mask='ff0')  # TODO: This writes t/h 3/4 but enable_transfer etc do not
            for num_of_acquisitions in xrange(self.noa):
                # def step():
                filename = self.path + '/{:0.3f}.out'.format(time.time())
                cmd = ['pci', '-r', 'dma0', '--multipacket', '-o', filename, '--timeout', str(self.timeout * 1000000)]
                # 1000000 because pci expects microseconds
                data_raw = board.safe_call(cmd)
                # rename with correct timestamp - last modified time
                newfile = '{path}/trigger_{num:05}_{htime}_{unixtime}.out'.format(num=num_of_acquisitions,
                                                                                  htime=dt.fromtimestamp(
                                                                                      os.path.getmtime(
                                                                                          filename)).strftime(
                                                                                      '%Y-%m-%dT%Hh%Mm%Ss%f'),
                                                                                  unixtime=int(
                                                                                      os.path.getmtime(filename)),
                                                                                  path=self.path)
                os.rename(filename, newfile)
                self.countUpdate.emit(num_of_acquisitions + 1)
                if os.path.getsize(newfile) > 0:
                    self.liveplot.emit(newfile)
                else:
                    logging.info("Acquired 0b, possible trigger timeout.")
                if self._quit:
                    break

            self.finished.emit()

        def wait_rw_seq(self):
            for num_of_acquisitions in xrange(self.noa):
                board.write_pci('00bf0', hex_mask='CF0')  # enable readout
                pre_acq_num = board.read_pci(1, '9034')[0]
                time_a = time.time()
                timeout = False
                while pre_acq_num == board.read_pci(1, '9034')[0]:
                    if time.time() - time_a > self.timeout:
                        timeout = True
                        break
                    if self._quit:
                        self.finished.emit()
                        return
                if not timeout:
                    board.write_pci('000f0', hex_mask='8F0')  # disable readout
                    #                    board.enable_transfer()
                    board.write_pci('007f0', hex_mask='CF0')  # enable transfer
                    filename = self.path + '/{:0.3f}.out'.format(time.time())
                    cmd = ['pci', '-r', 'dma0', '--multipacket', '-o', filename, '--timeout',
                           str(self.timeout * 1000000)]
                    # 1000000 because pci expects microseconds
                    data_raw = board.safe_call(cmd)
                    board.write_pci('000f0', hex_mask='4F0')  # disable transfer
                    newfile = '{path}/trigger_{num:05}_{htime}_{unixtime}.out'.format(num=num_of_acquisitions,
                                                                                      htime=dt.fromtimestamp(
                                                                                          os.path.getmtime(
                                                                                              filename)).strftime(
                                                                                          '%Y-%m-%dT%Hh%Mm%Ss%f'),
                                                                                      unixtime=int(
                                                                                          os.path.getmtime(filename)),
                                                                                      path=self.path)
                    os.rename(filename, newfile)
                    self.liveplot.emit(newfile)
                else:
                    logging.info("Trigger timeout.")

                self.countUpdate.emit(num_of_acquisitions + 1)
                if self._quit:
                    break

            self.finished.emit()

        def quit(self):
            self._quit = True

        def __del__(self):
            board.write_pci('0', '9024')
            time.sleep(0.1)
            # board.safe_call(['pci', '-w', '902C', str(skip)])
            board.write_pci('0', '902C')
            time.sleep(0.1)
            board.write_pci('3f0', hex_mask='ff0')  # TODO: This writes t/h 3/4 but enable_transfer etc do not

    def finished():
        global thread_wait
        # board.safe_call(['pci', '-w', '9024', str(num_of_acquisitions)])
        board.write_pci('0', '9024')
        time.sleep(0.1)
        # board.safe_call(['pci', '-w', '902C', str(skip)])
        board.write_pci('0', '902C')
        time.sleep(0.1)
        board.write_pci('3f0', hex_mask='ff0')  # TODO: This writes t/h 3/4 but enable_transfer etc do not

        Elements.setEnabled('acquire', True)

        thread_wait.quit()
        thread_wait.wait()
        thread_wait = None
        board.status.wait = False
        trigger_progressbar.remove(0)
        log(additional="Stop wait on trigger")
        for elem in Elements.getElements("acquireTrigger"):
            if isinstance(elem, QtGui.QShortcut):
                continue
            elem.setIcon(QtGui.QIcon(config.install_path + config.startIcon))
            elem.setText(tr("Button", "Start Acquisition"))
            elem.setEnabled(True)

        return

    global twt
    global thread_wait
    thread_wait = QtCore.QThread()
    twt = thread_wait_on_signal(num_of_acquisitions, storage.storage.save_location + '/' + storage.storage.subdirname,
                                timeout)
    twt.countUpdate.connect(trigger_progressbar.setValue)
    twt.finished.connect(lambda: finished())
    twt.liveplot.connect(_bif_read_and_update_data_from_file)
    twt.moveToThread(thread_wait)
    if method == 1:
        thread_wait.started.connect(twt.wait_rw_simul)
    elif method == 2:
        thread_wait.started.connect(twt.wait_rw_seq)
    else:
        raise ValueError("Wrong method")
    thread_wait.start()
