"""
This module defines the Initial view of the gui
"""

from PyQt4 import QtGui, QtCore
import logging

import kcgwidget as kcgw
from backend import board
from groupedelements import Checkboxes, Buttons, Elements
import backendinterface as bif
import storage
from .. import config

tr = kcgw.tr

class LED(QtGui.QWidget):
    """
    Produces a graphical LED
    """
    def __init__(self, parent=None, status=1, height=10, width=10):
        """
        Initialize a LED
        :param parent: (QWidget) parent of this widget
        :param status: (int) (0=out, 1=off, 2=orange, 3=on) initial status of this LED
        :param height: (int) height of the LED
        :param width: (int) width of the LED
        :return: -
        """
        QtGui.QWidget.__init__(self, parent)
        colorRGB=(255, 0, 0)
        self.width = width
        self.height = height
        self.color = QtGui.QColor(colorRGB[0], colorRGB[1], colorRGB[2])
        self.center = QtCore.QPoint(width, height)
        self.setMinimumSize(2 * width, 2 * height)
        self.setMaximumSize(2 * width, 2 * height)
        if status == 3:
            self.set_on()
        elif status == 1:
            self.set_off()
        elif status == 0:
            self.set_out()
        elif status == 2:
            self.set_tri()
        self.status = status

    def paintEvent(self, event):
        paint = QtGui.QPainter()
        paint.begin(self)
        paint.setRenderHint(QtGui.QPainter.Antialiasing)

        # draw a grey 'socket' for the LED
        paint.setPen(QtGui.QColor(160, 160, 160))
        paint.setBrush(QtGui.QColor(180, 180, 180))
        paint.drawEllipse(self.center, self.width, self.height)

        # draw the body of the LED
        paint.setBrush(self.color)
        paint.drawEllipse(self.center, self.width*0.85, self.height*0.85)

    def set_on(self):
        """
        Set the LED to "on" state
        :return: -
        """
        self.color = QtGui.QColor(0, 255, 0)
        self.update()
        self.status = 3
    def set_off(self):
        """
        Set the LED to "off" state
        :return: -
        """
        self.color = QtGui.QColor(255, 0, 0)
        self.update()
        self.status = 1
    def set_out(self):
        """
        Set the LED to "OUT" state (that is like an LED without power)
        :return: -
        """
        self.color = QtGui.QColor(150, 150, 150)
        self.update()
        self.status = 0
    def set_tri(self):
        """
        Set the LED to "TRI" state (that is led is orange)
        :return: -
        """
        self.color = QtGui.QColor(255, 255, 0)
        self.update()
        self.status = 2

    def set_status(self, status):
        """
        Set the status of the led
        :param status: status (in 0, 1, 2, 3)
        """
        if status == 0:
            self.set_out()
        elif status == 1:
            self.set_off()
        elif status == 2:
            self.set_tri()
        elif status == 3:
            self.set_on()

class StatusLED(QtGui.QWidget):
    """
    Create a Status LED with Label next to it
    """
    def __init__(self, text, status=None):
        """
        Initialise StatusLED
        :param text: label text next to the LED
        :param status: initial status of the LED
        :return: -
        """
        super(StatusLED, self).__init__()
        self.layout = QtGui.QHBoxLayout()
        self.label = QtGui.QLabel(text)
        self.led = LED(status=status, width=9, height=9)
        self.layout.addWidget(self.led)
        self.layout.addWidget(self.label)
        self.setLayout(self.layout)
    def set_on(self):
        """ See set_on of LED Class """
        self.led.set_on()
    def set_off(self):
        """ See set_off of LED Class """
        self.led.set_off()
    def set_out(self):
        """ See set_out of LED Class """
        self.led.set_out()
    def set_tri(self):
        """ See set_tri of LED Class """
        self.led.set_tri()
    def set_status(self, status):
        """ See set_status of LED Class """
        self.led.set_status(status)

class LogHandler(logging.Handler):
    """
    Handler Class to configure Logging
    This also enables logging to the output field in the gui
    """
    def __init__(self, log_area):
        logging.Handler.__init__(self)
        self.text_area = log_area
        self.setFormatter(logging.Formatter('%(asctime)s: %(message)s'))

    def emit(self, record):
        self.text_area.append(self.format(record))



class ControlWidget(kcgw.KCGWidgets):
    """
    Main Widget that is shown at start of gui.
    """
    def __init__(self):
        super(ControlWidget, self).__init__()

        self.initUi()
        storage.update_LED = self.on_check # register the led updater function (used in backendinterface.bk_status_readout)

    def initUi(self):
        """
        This Method will be called at creation of a ControlWidget instance (by __init__).
        It creates the various Widgets such as buttons etc. used to build the widget.
        :return:
        """

        self.overlayout = QtGui.QVBoxLayout()
        self.setLayout(self.overlayout)
        self.layout = QtGui.QHBoxLayout()
        self.overlayout.addLayout(self.layout)
        self.overlayout.addStretch(1)

        self.gridLayout = QtGui.QGridLayout()
        self.gridLayout.addWidget(self.createLabel(tr("Label", "Board Control")), 0, 0)
        self.gridLayout.addWidget(self.createLabel(tr("Label", "Detailed Board Control")), 2, 0)
        self.gridLayout.addWidget(self.createLabel(tr("Label", "Board Status")), 2, 1)

        # ----------[ LED Status ]---------------
        self.pipeline_led = StatusLED(tr("Label", "DataFlow Pipeline"))
        self.master_control_led = StatusLED(tr("Label", "Master Control"))
        self.data_check_led = StatusLED(tr("Label", "Data Check"))
        self.PLL_LD_led = StatusLED(tr("Label", "PLL_LD"))
        self.gridLayout.addWidget(self.pipeline_led, 3, 1)
        self.gridLayout.addWidget(self.master_control_led, 4, 1)
        self.gridLayout.addWidget(self.data_check_led, 5, 1)
        self.gridLayout.addWidget(self.PLL_LD_led, 6, 1)

        # -----------[ Buttons ]--------------
        self.all_in_one_button = self.createButton(text=tr("Button", "Prepare Board"), connect=self.all_in_one,
                                                   tooltip=tr("Tooltip", "Start, Calibrate, Synchronize and set Defaults\nCtrl+A"))
        self.all_in_one_button.setShortcut("Ctrl+A")
        self.all_in_one_button.setObjectName("all_in_one")

        self.skip_init_button = self.createButton(text=tr("Button", "Skip Initialisation"), connect=self.skip_init,
                                                  tooltip=tr("Tooltip", "Skip Initialisation and read values from board.\n"
                                                                        "NOTE: ONLY DO THIS IF BOARD WAS CALIBRATED AND SYNCHRONIZED BEFORE"))

        self.start_button = self.createButton(text=tr("Button", "Start Board"), connect=bif.bk_start_board)
        self.calibrate_button = self.createButton(text=tr("Button", "Calibrate"), connect=bif.bk_calibrate)
        self.syncronize_button = self.createButton(text=tr("Button", "Synchronize"), connect=bif.bk_sync_board)
        self.set_default_button = self.createButton(text=tr("Button", "Set Defaults"), connect=lambda: bif.bk_write_values(defaults=True))
        self.soft_reset_button = self.createButton(text=tr("Button", "Soft Reset"), connect=bif.bk_soft_reset)
        self.off_button = self.createButton(text=tr("Button", "Board Off"), connect=bif.bk_stop_board)
        self.off_button.setObjectName("off")
        self.check_status_button = self.createButton(text=tr("Button", "Check Status"), connect=bif.bk_status_readout)

        Buttons.addButton("start_board", self.start_button)
        Buttons.addButton(["calibrate", "after_start"], self.calibrate_button)
        Buttons.addButton(["synchronize", "after_start"], self.syncronize_button)
        Buttons.addButton(["set_defaults", "after_start"], self.set_default_button)
        Elements.addItem('no_board',
                         [
                             self.start_button,
                             self.calibrate_button,
                             self.syncronize_button,
                             self.set_default_button,
                             self.soft_reset_button,
                             self.off_button,
                             self.all_in_one_button
                         ])

        # -----------[ Add to grid ]---------------
        self.gridLayout.addWidget(self.all_in_one_button, 1, 0)
        self.gridLayout.addWidget(self.skip_init_button, 1, 1)
        self.gridLayout.addWidget(self.start_button, 3, 0)
        self.gridLayout.addWidget(self.calibrate_button, 4, 0)
        self.gridLayout.addWidget(self.syncronize_button, 5, 0)
        self.gridLayout.addWidget(self.set_default_button, 6, 0)
        self.gridLayout.addWidget(self.soft_reset_button, 7, 0)
        self.gridLayout.addItem(QtGui.QSpacerItem(10, 50), 8, 0)
        self.gridLayout.addWidget(self.off_button, 9, 0)
        self.gridLayout.addWidget(self.check_status_button, 7, 1)


        self.layout.addLayout(self.gridLayout)
        self.layout.addStretch(1)
        self.layout.addWidget(self.createLabel(image=QtGui.QPixmap(config.install_path+"style/KAPTURE.jpg")))
        self.log_area = QtGui.QTextEdit()
        self.log_area.setReadOnly(True)
        self.overlayout.addWidget(self.log_area)

        self.setTabOrder(self.soft_reset_button, self.check_status_button)
        self.setTabOrder(self.check_status_button, self.off_button)

        # ------------------[ Logging ]----------------------
        log_handler = LogHandler(self.log_area)
        self.logger = logging.getLogger()
        board.config.register_logger(self.logger)
        self.logger.addHandler(log_handler)
        self.logger.setLevel(logging.INFO)
        logging.logger = self.logger
        self.gridLayout.addWidget(self.all_in_one_button, 1, 0)


    def all_in_one(self):
        """
        Function that gets called when the Prepare Board Button is pressed.
        It Starts the board, syncs it and sets defaults.
        This is accomplished via the backendinterface module
        :return: -
        """
        bif.bk_start_board()
        def do_the_rest():
            bif.bk_sync_board()
            bif.bk_write_values(defaults=True)
        bif.bk_calibrate(do_the_rest)

    def skip_init(self):
        board.config.read_from_board()
        board.status.calibrated = True
        board.status.synced = True
        board.status.defaults_set = True
        bif.bk_status_readout()

    def on_check(self):
        """
        This function is the handler for the status leds on the ControlWidget View.
        Parses the registers and sets the colors of the leds according.
        :param registers: The registers read from board.
        :return: -
        """
        try:
            status = bif.bk_get_status()
        except IndexError:
            QtGui.QMessageBox.critical(self, "Status Update Error", "Not enough registers given for status update.")
            return
        for led, st in status.items():
            getattr(self, led+"_led").set_status(st)