"""
This is the container widget for multiple subwindows
"""
from PyQt4 import QtGui, QtCore

import kcgwidget as kcgw
from leftbar import LeftBar
import leftbar
from .. import config

tr = kcgw.tr

class WidgetTypeError(Exception):
    """
    Simple error that describes when a wrong window type gets added
    """
    pass

class MDIArea(kcgw.KCGWidgets):
    """
    The MDI Area used by Multiwidget
    """
    def __init__(self, parent):
        super(MDIArea, self).__init__("MDIArea")
        self.parent = parent

        self.widgets = {}
        self.area = QtGui.QMdiArea()

    def newWidget(self, widget, name, unique_id, widget_type, minSize=False):
        """
        Add a new Widget to the MDIArea
        :param widget: (subclass of QMdiSubWindow) The widget to show
        :param name: (str) name of the window
        :param unique_id: (int) unique id of the window
        :param widget_type: (int) the type of this window
        :param minSize: (bool) whether to shrink the window to minimum size upon creation
        :return: -
        """
        if not isinstance(widget, kcgw.KCGWidgets):
            raise WidgetTypeError("Newly Created Widget is of type " +
                                  type(widget).__name__ +
                                  ". Expected class derived of " +
                                  kcgw.KCGWidgets.__name__ + ".")
        subWidget = kcgw.KCGSubWidget(name=name, unique_id=unique_id, typ=widget_type, minSize=minSize)
        subWidget.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        widget.setParent(subWidget)

        if minSize:
            subWidget.resize(widget.minimumSizeHint()*1.2)


        widget.closeSignal.connect(subWidget.close)
        subWidget.setWidget(widget)
        self.area.addSubWindow(subWidget)
        subWidget.show()
        self.widgets[unique_id] = widget

class MultiWidget(QtGui.QWidget):
    """
    The Widget used as Multiwidget. This is the main View during operation with KCG.
    """
    def __init__(self):
        super(MultiWidget, self).__init__()
        self.layout = QtGui.QVBoxLayout()
        self.setLayout(self.layout)

        self.splitter = QtGui.QSplitter(QtCore.Qt.Horizontal)

        self.leftBar = LeftBar(self)
        self.area = MDIArea(self)
        self.toolbar = QtGui.QToolBar()
        kcgw.area = self.area

        # --------[ ToolBar ]--------
        self.new_live_action = QtGui.QAction(QtGui.QIcon(config.install_path + config.newPlotLiveIcon), tr("Button", "New Live Plot"), self)
        self.new_live_action.setShortcut("Ctrl+L")
        self.toolbar.addAction(self.new_live_action)
        self.new_live_action.triggered.connect(lambda: self.leftBar.add_plot(leftbar.LIVE))
        self.new_live_action.setToolTip(tr("Button", "New Live Plot") + "\n(Ctrl+L)")

        self.new_data_action = QtGui.QAction(QtGui.QIcon(config.install_path + config.newPlotDataIcon), tr("Button", "New Data Plot"), self)
        self.new_data_action.setShortcut("Ctrl+D")
        self.toolbar.addAction(self.new_data_action)
        self.new_data_action.triggered.connect(lambda: self.leftBar.add_plot(leftbar.FILE))
        self.new_data_action.setToolTip(tr("Button", "New Data Plot") + "\n(Ctrl+L)")

        self.layout.addWidget(self.toolbar)
        # ------[ End Tool Bar ]-----

        self.splitter.addWidget(self.leftBar)
        self.splitter.addWidget(self.area.area)
        self.layout.addWidget(self.splitter)

        self.evaluate_registered_widgets()
        self.evaluate_registered_widget_functions()

    def addToolbarButton(self, icon, text, target, shortcut=None):
        """
        Add a toolbar button.
        :param icon: (QIcon) The icon to show on the toolbar button
        :param text: (str) tooltip for this button
        :param target: (callable) The function to call upon press on button
        :param shortcut: (str) Keyboard shortcut to call target
        :return: -
        """
        na = QtGui.QAction(icon, text, self)
        if shortcut:
            na.setShortcut(shortcut)
            na.setToolTip(text + "\n("+shortcut+")")
        self.toolbar.addAction(na)
        na.triggered.connect(target)

    def evaluate_registered_widgets(self):
        """
        Evaluate all the registered widgets and add a toolbar button for those
        :return:
        """
        for f in kcgw.get_registered_widgets():
            self.addToolbarButton(*f)

    def evaluate_registered_widget_functions(self):
        """
        Evaluate the functions that are registered
        Those are in general functions that have to be called after widgets are created
        :return:
        """
        for f in kcgw.get_registered_widget_functions():
            f()

