"""
This Module implements the Plot Windows used in KCG
"""
import pyqtgraph as pg
pg.setConfigOption('useOpenGl', True)
pg.setConfigOption('background', 'w')
pg.setConfigOption('foreground', 'k')
from threading import Thread
import numpy as np
from PyQt4 import QtCore, QtGui
import logging
from groupedelements import live_plot_windows
from backend import io
from backend import board

import kcgwidget as kcgw
tr = kcgw.tr
# from backend.dataset import DataSet
# from backend import io
# import storage

LIVE = 1
FILE = 2
ERROR = 99
NO_DATA = 98

class Enum():
    """
    Simple Enum Class (as this is not supported by clean python)
    """
    def __init__(self, *args):
        self.idx = 0
        for i in args:
            setattr(self, i, self.idx)
            self.idx += 1

plotList = [tr("Label", "Heatmap"), tr("Label", "FFT"), tr("Label", "Trains"), tr("Label", "Combined"), tr("Label", "Compare")]
PlotType = Enum(*[str(i) for i in plotList])

class SpectrogramColorLegendItem(pg.GraphicsWidget):
    """
    The Item used as Legend for Heatmap and FFT Plot
    """
    gradientChanged = QtCore.pyqtSignal()
    def __init__(self, img=None):
        """
        Initialise the Legend
        :param img: (ImageItem) the image item this is the legend for
        :return: -
        """
        super(SpectrogramColorLegendItem, self).__init__()
        self.layout = QtGui.QGraphicsGridLayout()
        self.setLayout(self.layout)
        self.img = img
        self.gei = pg.GradientEditorItem(orientation='right')
        self.legend_axis = pg.AxisItem(orientation='left')
        # self.addItem(self.legend_axis)
        self.layout.addItem(self.legend_axis, 0, 0, alignment=QtCore.Qt.AlignVCenter)
        # self.addItem(self.gei)
        self.layout.addItem(self.gei, 0, 1)
        self.gei.sigGradientChanged.connect(self.gradient_changed)
        self.gei.loadPreset('spectrum')
        # self.setFixedWidth(80)
        self.image_changed()

    def gradient_changed(self):
        """
        Proxy function for the gradient_cnahged event of the GradientEditorItem
        :return:
        """
        self.gradientChanged.emit()

    def set_image(self, img):
        """
        Set the Image this is the legend for (only needed if not already done upon initialisation)
        :param img: (ImageItem) The image item to use
        :return: -
        """
        self.img = img
        self.image_changed(True)

    def update_axis(self):
        """
        Update the axis of this legend
        :return: -
        """
        if self.img.getLevels():
            min, max = self.img.getLevels()
            self.legend_axis.setRange(min, max)

    def reset_gradient(self):
        """
        Reset the gradient to the preset "spectrum"
        :return: -
        """
        self.gei.loadPreset('spectrum')

    def image_changed(self, reset_gradient=True):
        """
        Call this to adjust the legend when the image has changed
        :param reset_gradient: (bool) whether to reset the gradient or not
        :return:
        """
        if self.img:
            self.update_axis()
            if reset_gradient:
                self.reset_gradient()

    def getLookupTable(self, nPts, alpha=None):
        """
        Get the look up table of the imageitem
        :param nPts: (int) number of points the lookup table is defined on
        :param alpha: ??
        :return: -
        """
        return self.gei.getLookupTable(nPts, alpha=alpha)

    def resizeEvent(self, event):
        """
        Handle resizing of the window
        """
        self.legend_axis.setHeight(self.gei.length)

class SubPlotWidget(pg.GraphicsLayoutWidget):
    """
    The Widget actually containing the plots and images
    """
    def __init__(self, dType=FILE):
        super(SubPlotWidget, self).__init__()

        self.dType = dType

        # self.setStyleSheet("border: 5px solid black; margin: 10px;")
        self.plotType = 1
        self.plotItem = pg.PlotItem()
        self.img = pg.ImageItem()
        self.plotItem2 = pg.PlotItem()
        self.img2 = pg.ImageItem()
        self.plotItem2.addItem(self.img2)

        self.gradient_legend = SpectrogramColorLegendItem(self.img)
        def changelut():
            """ Handle changing of the lookup table"""
            self.img.setLookupTable(self.gradient_legend.getLookupTable(512))
            self.img2.setLookupTable(self.gradient_legend.getLookupTable(512))
        self.gradient_legend.gradientChanged.connect(changelut)

        self.addItem(self.plotItem)
        self.plotItem.addItem(self.img)
        self.plotItemPlotScatter = pg.PlotDataItem(pen=None, symbolPen=pg.mkPen(50, 50, 255), symbolSize=3, pxMode=True, symbolBrush=pg.mkBrush(50, 50, 255), downsample=200)
        self.plotItemPlotScatter1 = pg.PlotDataItem(pen=None, symbolPen=pg.mkPen(255, 50, 50), symbolSize=3, pxMode=True, symbolBrush=pg.mkBrush(255, 50, 50), downsample=200)
        # self.plotItemPlotScatter2 = pg.PlotDataItem(pen=None, symbolPen=pg.mkPen(50, 255, 50), symbolSize=3, pxMode=True, symbolBrush=pg.mkBrush(50, 255, 50), downsample=200)
        self.plotItemPlot = pg.PlotDataItem()  # pen="#000000")
        self.plotItem.addItem(self.plotItemPlot)
        self.plotItem.addItem(self.plotItemPlotScatter)
        self.plotItem.addItem(self.plotItemPlotScatter1)
        # self.plotItem.addItem(self.plotItemPlotScatter2)

        self.plotItem.vb.origAutoRange = self.plotItem.vb.autoRange

        self.error_label = pg.LabelItem("Error during plot.", color=(200, 50, 50))
        self.error_label.hide()
        self.error_label.scale(1, -0.6)
        self.error_label.translate(0, -30)
        self.plotItem.addItem(self.error_label)

        self.no_data_label = pg.LabelItem("No data to plot.", color=(200, 50, 50))
        self.no_data_label.hide()
        self.no_data_label.scale(1, -0.6)
        self.no_data_label.translate(0, -30)
        self.plotItem.addItem(self.no_data_label)


        # pos = np.array([0, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 1])
        # color = np.array([[0, 0, 255, 255], [0, 255, 255, 255], [0, 255, 0, 255], [130, 255, 0, 255], [255, 255, 0, 255], [255, 180, 0, 255], [255, 100, 0, 255], [255, 0, 0, 255]])
        # cmap = pg.ColorMap(pos, color)
        # self.lut = cmap.getLookupTable(0.0, 1.0, 256)
        # self.img.setLookupTable(self.lut)

    def _enableCustomAutoRange(self, data):
        """
        Enable custom auto range in this plot
        :param data: the data to autorange
        :return:
        """
        def newAutoRange(*args, **kwargs):
            """ function to handle the new autorange """
            bounds = [np.min(data), np.max(data)]
            self.plotItem.vb.setRange(xRange=[0, len(data)],
                yRange=[bounds[0]-0.1*(bounds[1]-bounds[0])-1, bounds[1]+0.1*(bounds[1]-bounds[0])+1], update=True)
            self.plotItem.update()
        self.plotItem.vb.autoRange = newAutoRange
        self.plotItem.autoBtn.clicked.disconnect()
        self.plotItem.autoBtn.clicked.connect(newAutoRange)

    def _disableCustomAutoRange(self):
        """
        Disable the custom autorange and reset it to default
        :return: -
        """
        self.plotItem.vb.autoRange = self.plotItem.vb.origAutoRange
        self.plotItem.autoBtn.clicked.disconnect()
        self.plotItem.autoBtn.clicked.connect(self.plotItem.autoBtnClicked)


    @QtCore.pyqtSlot(np.ndarray, tuple, tuple)
    def plot(self, data, xvalueborders=None, yvalueborders=None, autorange=True):
        """
        Plot Data. The plot type depends on the type property
        :param data: (dataset.DataSet) data to plot
        :param xvalueborders: (touple) the borders for the xvalues
        :param yvalueborders: (touple) the borders for the yvalues
        :param autorange: (bool) whether to perform a autorange or not
        :return: -
        """
        self.error_label.hide()
        self.no_data_label.hide()
        self.plotItemPlot.clear()
        self.plotItemPlotScatter.clear()
        self.plotItemPlotScatter1.clear()
        # self.plotItemPlotScatter2.clear()
        self.plotItem.resetTransform()
        self.plotItem.getAxis('bottom').setScale()
        try:
            self.removeItem(self.plotItem2)
        except Exception as e:
            if "Could not determine index of item" in str(e):
                pass
            else:
                raise

        if self.plotType == PlotType.FFT or self.plotType == PlotType.Heatmap:
            self.addItem(self.gradient_legend)
            self.gradient_legend.show()

            # self.parent().max_slider.setMinimum(np.min(data))
            # self.parent().max_slider.setMaximum(np.max(data))
            # self.parent().low_slider.setMinimum(np.min(data))
            # self.parent().low_slider.setMaximum(np.max(data))


            self.img.setImage(data, scale=[1, 1/1000.])
            self.img.resetTransform()
            # self.parent().max_slider.setValue(self.img.getLevels()[1])
            # self.parent().low_slider.setValue(self.img.getLevels()[0])
            self.img.setAutoDownsample(True)
            if autorange:
                self.gradient_legend.image_changed()
                pass

            if xvalueborders:
                self.plotItem.getAxis('bottom').setScale((xvalueborders[1]-xvalueborders[0])/float(data.shape[0]))
                self.img.translate(xvalueborders[0]/self.plotItem.getAxis('bottom').scale, 0)

            else:
                self.plotItem.getAxis('bottom').setScale()

            if yvalueborders:
                self.img.translate(0, yvalueborders[0])
                self.plotItem.getAxis('left').setScale(yvalueborders[1]/float(data.shape[1]))

            # if self.dType == FILE:
            self._disableCustomAutoRange()
            self.plotItem.setClipToView(True)

        if self.plotType == PlotType.Compare:
            self.addItem(self.plotItem2)
            self.plotItem2.vb.linkView(self.plotItem2.vb.XAxis, self.plotItem.vb)
            self.plotItem2.vb.linkView(self.plotItem2.vb.YAxis, self.plotItem.vb)
            self.addItem(self.gradient_legend)
            self.gradient_legend.show()

            # self.parent().max_slider.setMinimum(np.min(data))
            # self.parent().max_slider.setMaximum(np.max(data))
            # self.parent().low_slider.setMinimum(np.min(data))
            # self.parent().low_slider.setMaximum(np.max(data))


            self.img2.setImage(data[1], scale=[1, 1/1000.])
            self.img2.show()
            self.img.setImage(data[0], scale=[1, 1/1000.])
            self.img.resetTransform()
            self.img2.resetTransform()
            # self.parent().max_slider.setValue(self.img.getLevels()[1])
            # self.parent().low_slider.setValue(self.img.getLevels()[0])
            self.img.setAutoDownsample(True)
            self.img2.setAutoDownsample(True)
            if autorange:
                self.gradient_legend.image_changed()
                pass

            if xvalueborders:
                self.plotItem.getAxis('bottom').setScale((xvalueborders[1]-xvalueborders[0])/float(data[0].shape[0]))
                self.img.translate(xvalueborders[0]/self.plotItem.getAxis('bottom').scale, 0)
                self.img2.translate(xvalueborders[0]/self.plotItem.getAxis('bottom').scale, 0)

            else:
                self.plotItem.getAxis('bottom').setScale()

            if yvalueborders:
                self.img.translate(0, yvalueborders[0])
                self.img2.translate(0, yvalueborders[0])
                self.plotItem.getAxis('left').setScale(yvalueborders[1]/float(data[0].shape[1]))

            if self.dType == FILE:
                self._disableCustomAutoRange()
            self.plotItem.setClipToView(True)


        if self.plotType == PlotType.Trains:
            self.img.clear()
            self.gradient_legend.hide()
            try:
                self.removeItem(self.gradient_legend)
                pass
            except Exception, e:
                if not "Could not determine index of item" in str(e):
                    raise
            self.plotItemPlotScatter.clear()
            self.plotItemPlotScatter1.clear()
            # self.plotItemPlotScatter2.clear()
            self.plotItemPlot.setData(data)

            # if self.dType == FILE:
            self._enableCustomAutoRange(data)


        if self.plotType == PlotType.Combined:
            self.img.clear()
            self.gradient_legend.hide()
            try:
                self.removeItem(self.gradient_legend)
                pass
            except Exception, e:
                if not "Could not determine index of item" in str(e):
                    raise
            self.plotItemPlot.clear()
            self.plotItemPlotScatter.setData(data[0].transpose())
            self.plotItemPlotScatter1.setData(data[1].transpose())

            if self.dType == FILE:
                self._enableCustomAutoRange(data[0][1])

            self.plotItem.setClipToView(False) # NOTE: otherwise only a very small portion of data is visible :wonder:
        if self.plotType == ERROR or self.plotType == NO_DATA:
            self.img.hide()
            self.gradient_legend.hide()
            self.img2.hide()
            self.plotItemPlot.hide()
            self.plotItemPlotScatter.hide()
            self.plotItemPlotScatter1.hide()
            if self.plotType == ERROR:
                self.error_label.show()
            else:
                self.no_data_label.show()
            # self.addItem(self.error_label)
            # self.removeItem(self.plotItem)

        if autorange:
            # if self.plotType in [PlotType.FFT, PlotType.Heatmap, PlotType.Trains]:
            self.plotItem.getViewBox().autoRange()
            self.plotItem.getViewBox().autoRange()

        self.labelAxes()

    def labelAxes(self):
        """
        Add Labels to the axis depending on the self.plotType property.
        :return: -
        """
        if self.plotType == PlotType.Heatmap:
            self.plotItem.setLabel('bottom', 'Turn', '')
            self.plotItem.setLabel('left', 'Bunch Position', '')
        elif self.plotType == PlotType.FFT:
            self.plotItem.setLabel('left', 'Bunch Position', '')
            self.plotItem.setLabel('bottom', 'Frequency', 'Hz')
        elif self.plotType == PlotType.Trains:
            self.plotItem.setLabel('left', '', '')
            self.plotItem.setLabel('bottom', 'Sample Point', '')
        elif self.plotType == PlotType.Combined:
            self.plotItem.setLabel('left', '', '')
            self.plotItem.setLabel('bottom', '', '')
        if self.plotType == PlotType.Compare:
            self.plotItem.setLabel('bottom', 'Turn', '')
            self.plotItem.setLabel('left', 'Bunch Position', '')
            self.plotItem2.setLabel('bottom', 'Turn', '')
            self.plotItem2.setLabel('left', 'Bunch Position', '')
            # self.plotItem2.hideAxis('left')




    def changeType(self, type):
        """
        Change the plot Type
        :param type: (int) the new type
        :return: -
        """
        self.plotType = type
        self.gradient_legend.image_changed()

class PlotWidget(kcgw.KCGWidgets):
    """
    The container Class holding various buttons and controls and the actual plots as SubPlotWidgets instance
    """
    close_signal = QtCore.pyqtSignal()
    change_type_signal = QtCore.pyqtSignal(int, int, str)
    def __init__(self, parent=None, name=None, unique_id=None, type=None, datatype=None, prefix=None, fName=None, data=None):
        """
        Initialise the Plot Widgt
        :param parent: (QWidget) the parent widget
        :param name: (str) name of this widget
        :param unique_id: (int) unique id of this widget
        :param type: (int) type of this widget (not plot type)
        :param datatype: (int) datatype (LIVE or FILE)
        :param prefix: (str) prefix for text in the listview in LeftBar
        :param fName: (str) the filename (this is only used if datatype is FILE)
        :param data: (dataset.DataSet) the data to be plotted
        :return: -
        """
        super(PlotWidget, self).__init__()
        if name != None:
            self.theName = name
        if unique_id != None:
            self.theId = unique_id
        if type != None:
            self.theType = type
        if datatype != None:
            self.theDataType = datatype
        if prefix != None:
            self.thePrefix = prefix
        if self.theDataType == FILE:
            self.fName = fName
        if self.theDataType == FILE:
            self.data = data
        else:
            self.data = None

        self.close_silent = False
        self.parent = parent
        self.adc = 1
        self.secadc = 2
        self._single_adc_checked = False
        self.initUI()
        self.changePlotType(self.theType)  # initially mark the correct button
        self.setWindowTitle(self.theName + " - " + plotList[type] + " - " + str(self.thePrefix))
        self.plot_widget.changeType(type)
        if self.theDataType == FILE:
            self.plot(type)
        else:
            live_plot_windows.addWindow(self)
            if board.status.last_file is not None:
                self.data = io.read_from_file(board.status.last_file)
                self.plot(type)

    def initUI(self):
        """
        Initialise the UI
        :return: -
        """
        self.plot_widget = SubPlotWidget(dType=self.theDataType)
        self.layout = QtGui.QVBoxLayout()
        self.plot_buttons_layout = QtGui.QHBoxLayout()

        self.heatmap_button = self.createButton(text=tr("Button", "Heatmap"), connect=lambda: self.plot(type=PlotType.Heatmap))
        self.fft_button = self.createButton(text=tr("Button", "FFT"), connect=lambda: self.plot(type=PlotType.FFT))
        self.trains_button = self.createButton(text=tr("Button", "Trains"), connect=lambda: self.plot(type=PlotType.Trains))
        self.combined_button = self.createButton(text=tr("Button", "Combined"), connect=lambda: self.plot(type=PlotType.Combined))
        self.compare_button = self.createButton(text=tr("Button", "Compare"), connect=lambda: self.plot(type=PlotType.Compare))
        self.type_buttons = {PlotType.Heatmap:self.heatmap_button, PlotType.FFT:self.fft_button,
                             PlotType.Trains:self.trains_button, PlotType.Combined:self.combined_button,
                             PlotType.Compare:self.compare_button}
        self.defaultButtonStyleSheet = self.heatmap_button.styleSheet()

        self.adc1 = self.createCheckbox(text="ADC 1", connect=self.change_adc)
        self.adc1.setChecked(True)
        self.adc2 = self.createCheckbox(text="ADC 2", connect=self.change_adc)
        self.adc3 = self.createCheckbox(text="ADC 3", connect=self.change_adc)
        self.adc4 = self.createCheckbox(text="ADC 4", connect=self.change_adc)

        self.plot_buttons_layout.addWidget(self.heatmap_button)
        self.plot_buttons_layout.addWidget(self.fft_button)
        self.plot_buttons_layout.addWidget(self.trains_button)
        self.plot_buttons_layout.addWidget(self.combined_button)
        self.plot_buttons_layout.addWidget(self.compare_button)

        self.adc_checkbox_layout = QtGui.QHBoxLayout()
        self.adc_checkbox_layout.addWidget(self.adc1)
        self.adc_checkbox_layout.addWidget(self.adc2)
        self.adc_checkbox_layout.addWidget(self.adc3)
        self.adc_checkbox_layout.addWidget(self.adc4)

        self.groupWidget = QtGui.QWidget()
        self.groupWidget.setLayout(self.adc_checkbox_layout)
        self.group = QtGui.QButtonGroup()
        self.group.addButton(self.adc1, 1)
        self.group.addButton(self.adc2, 2)
        self.group.addButton(self.adc3, 3)
        self.group.addButton(self.adc4, 4)
        self.group.setExclusive(True)


        self.compare_heading_left = self.createLabel("ADC 1")
        self.compare_heading_left.setAlignment(QtCore.Qt.AlignCenter)
        self.compare_heading_right = self.createLabel("ADC 2")
        self.compare_heading_right.setAlignment(QtCore.Qt.AlignCenter)
        self.compare_heading = QtGui.QWidget()
        self.compare_heading_layout = QtGui.QHBoxLayout()
        self.compare_heading_layout.addWidget(self.compare_heading_left)
        self.compare_heading_layout.addWidget(self.compare_heading_right)
        self.compare_heading.setLayout(self.compare_heading_layout)
        self.compare_heading.hide()
        self.adc1Compare = self.createCheckbox(text="ADC 1", connect=lambda: self.change_adc_compare(who=1))
        self.adc2Compare = self.createCheckbox(text="ADC 2", connect=lambda: self.change_adc_compare(who=2))
        self.adc3Compare = self.createCheckbox(text="ADC 3", connect=lambda: self.change_adc_compare(who=3))
        self.adc4Compare = self.createCheckbox(text="ADC 4", connect=lambda: self.change_adc_compare(who=4))
        self.adc1Compare.setChecked(True)
        self.adc2Compare.setChecked(True)

        self.adc_checkbox_compare_layout = QtGui.QHBoxLayout()
        self.adc_checkbox_compare_layout.addWidget(self.adc1Compare)
        self.adc_checkbox_compare_layout.addWidget(self.adc2Compare)
        self.adc_checkbox_compare_layout.addWidget(self.adc3Compare)
        self.adc_checkbox_compare_layout.addWidget(self.adc4Compare)

        self.groupWidgetCompare = QtGui.QWidget()
        self.groupWidgetCompare.setLayout(self.adc_checkbox_compare_layout)
        self.groupCompare = QtGui.QButtonGroup()
        self.groupCompare.setExclusive(False)
        self.groupCompare.addButton(self.adc1Compare, 1)
        self.groupCompare.addButton(self.adc2Compare, 2)
        self.groupCompare.addButton(self.adc3Compare, 3)
        self.groupCompare.addButton(self.adc4Compare, 4)
        self.groupWidgetCompare.hide()


        self.from_to_layout = QtGui.QHBoxLayout()
        self.from_spinbox = self.createSpinbox(0, 100000000, interval=100, connect=lambda: self.plot(self.theType))
        self.to_spinbox = self.createSpinbox(0, 100000000, start_value=1000, interval=100, connect=lambda: self.plot(self.theType))
        self.from_to_layout.addStretch()
        self.from_to_layout.addWidget(self.createLabel(tr("Label", "From:")))
        self.from_to_layout.addWidget(self.from_spinbox)
        self.from_to_layout.addWidget(self.createLabel(tr("Label", "To:")))
        self.from_to_layout.addWidget(self.to_spinbox)

        # self.slider_layout = QtGui.QHBoxLayout()
        # self.low_slider = QtGui.QSlider(QtCore.Qt.Horizontal)
        # self.low_slider.setMinimum(0)
        # self.low_slider.setMaximum(1000)
        # self.low_slider.valueChanged.connect(self.low_changed)
        # self.max_slider = QtGui.QSlider(QtCore.Qt.Horizontal)
        # self.max_slider.setMinimum(100)
        # self.max_slider.setMaximum(10000)
        # self.max_slider.valueChanged.connect(self.max_changed)
        # self.slider_layout.addWidget(self.low_slider)
        # self.slider_layout.addWidget(self.max_slider)
        self.layout.addLayout(self.plot_buttons_layout)
        self.layout.addWidget(self.compare_heading)
        self.layout.addWidget(self.plot_widget)
        self.layout.addWidget(self.groupWidget)
        self.layout.addWidget(self.groupWidgetCompare)
        self.layout.addLayout(self.from_to_layout)
        # self.layout.addLayout(self.slider_layout)
        self.setLayout(self.layout)

    def change_adc(self):
        """
        Change the adc for which data is plotted
        :return: -
        """
        self.adc = self.group.checkedId()
        self.change_type_signal.emit(self.theId, self.theType, str(self.adc))
        self.plot(self.theType)

    def change_adc_compare(self, who):
        """
        Change the adcs displayed in a compare plot
        :return:
        """
        if not self._single_adc_checked:
            self.adc1Compare.setChecked(False)
            self.adc2Compare.setChecked(False)
            self.adc3Compare.setChecked(False)
            self.adc4Compare.setChecked(False)
            getattr(self, 'adc'+str(who)+'Compare').setChecked(True)
            self._single_adc_checked = True
        else:
            checked = [i.isChecked() for i in self.groupCompare.buttons()]
            self._single_adc_checked = False
            self.adc = np.where(np.array(checked) == True)[0][0] + 1 # +1 because adcs are 1 based and indices 0
            self.secadc = np.where(np.array(checked) == True)[0][1] + 1 # +1 because adcs are 1 based and indices 0
            self.compare_heading_left.setText("ADC "+str(self.adc))
            self.compare_heading_right.setText("ADC "+str(self.secadc))
            self.change_type_signal.emit(self.theId, self.theType, str(self.adc)+"+"+str(self.secadc))
            self.plot(self.theType)

    def disable_buttons(self, b_bool):
        """
        Disable the buttons on this widget
        This is not used at the moment
        :param b_bool: (bool) disable(False) or enable(True)
        :return: -
        """
        self.heatmap_button.setDisabled(b_bool)
        self.fft_button.setDisabled(b_bool)
        self.trains_button.setDisabled(b_bool)
        self.combined_button.setDisabled(b_bool)

    def changePlotType(self, type):
        self.theType = type
        for btype, button in self.type_buttons.iteritems():
            style = ""
            if btype == type:
                # button.setStyleSheet("QPushButton {background-color:lightgreen;}")
                style += "QPushButton {background-color:lightgreen;}"
            else:
                # button.setStyleSheet(self.defaultButtonStyleSheet)
                style += self.defaultButtonStyleSheet
            style += "QPushButton:focus{background-color: lightgrey!important; border-color: lightblue;}"  # Note: this does not work
            button.setStyleSheet(style)
    def plot(self, type=None):
        """
        Wrapper function to call the correct plot function depending on type
        :param type: (int) the plot type
        :return: -
        """
        try:
            if type == None:
                type = self.theType
            self.do_plot(type)
        except Exception, e:
            # QtGui.QMessageBox.critical(self, tr("Heading", "Error Plotting Data"), tr("Dialog", "There was an error plotting the data, make sure it is valid and try again."))
            # mb.exec_()
            if self.data is None or len(self.data.array) == 0:
                self.plot_widget.changeType(NO_DATA)
            else:
                self.plot_widget.changeType(ERROR)
            self.plot_widget.plot(None)

    def plot_live(self, type=None, data=None):
        """
        Function to call when livedata is to be plotted
        :param type: (int) plot type
        :param data: (dataset.DataSet) data to plot
        :return: -
        """
        self.data = data
        try:
            if type == None:
                type = self.theType
            if type == PlotType.Trains or type == PlotType.Combined:
                self.do_plot(type, autorange=True)
            else:
                self.do_plot(type, autorange=False)
        except Exception, e:
            if len(self.data.array) == 0:
                self.plot_widget.changeType(NO_DATA)
            self.plot_widget.changeType(ERROR)
            self.plot_widget.plot(None)

    def do_plot(self, type, autorange=True):
        """
        Actually perform a plot (this calls SubPlotWidget.plot)
        :param type: (int) plot type
        :param autorange: (bool) whether to perform a autorange upon plot
        :return: -
        """
        if self.theType != type:
            if type == PlotType.Combined:
                self.groupWidget.hide()
            elif self.theType == PlotType.Combined:
                self.groupWidget.show()

            if type == PlotType.Compare:
                self.groupWidget.hide()
                self.groupWidgetCompare.show()
                self.compare_heading.show()
            else:
                self.groupWidget.show()
                self.groupWidgetCompare.hide()
                self.compare_heading.hide()

            self.changePlotType(type)
            self.setWindowTitle(self.theName + " - " + plotList[type] + " - " + str(self.thePrefix))
            self.change_type_signal.emit(self.theId, type, str(self.adc))
            if self.theType == PlotType.Compare:
                self.change_type_signal.emit(self.theId, type, str(self.adc)+"+"+str(self.secadc))
            self.plot_widget.changeType(type)


        if self.theType == PlotType.Compare:
            self.groupWidget.hide()
            self.groupWidgetCompare.show()
            self.compare_heading.show()
        else:
            self.groupWidgetCompare.hide()
            self.compare_heading.hide()

        f = self.from_spinbox.value()
        t = self.to_spinbox.value() if self.to_spinbox.value() > f else f+1
        if type == PlotType.FFT:
            self.plot_widget.plot(np.abs(self.data.fft(adc=self.adc, frm=f, to=t+1, drop_first_bin=True)).transpose(),
                                  autorange=autorange,
                                  xvalueborders=[self.data.fft_freq_dist(), self.data.fft_max_freq()])
        if type == PlotType.Heatmap:
            self.plot_widget.plot(self.data.heatmap(adc=self.adc, frm=f, to=t).transpose(), autorange=autorange)
        if type == PlotType.Compare:
            self.plot_widget.plot([self.data.heatmap(adc=self.adc, frm=f, to=t).transpose(),self.data.heatmap(adc=self.secadc, frm=f, to=t).transpose()], autorange=autorange)
        if type == PlotType.Trains:
            self.plot_widget.plot(self.data.train(adc=self.adc, frm=f, to=t), autorange=autorange)
        if type == PlotType.Combined:
            self.plot_widget.plot(self.data.combined(frm=f, to=t), autorange=autorange)

    def closeEvent(self, event):
        """
        Event Handler to handle the event of closing this window
        :param event: QEvent
        :return: -
        """
        if not self.close_silent:
            if not self.parent.remove_plot(self.theId, silent=self.close_silent):
                event.ignore()
                return
        del self.data
        del self.plot_widget
        super(PlotWidget, self).closeEvent(event)
        if self.theDataType == LIVE:
            live_plot_windows.removeWindow(self)
        self.close_signal.emit()
        del self
